#include "audio.h"
#include "video.h"
#include "title.h"
#include "group.h"
#include "tools.h"
#include "project.h"
#include "avswrapper.h"
#include "config.h"

#include "safefgets/safefgets.h"

Audio::Audio(Video * video, Audio * audio) : ManagablePtr()
{
	m_Video = video;
	if (audio) {
		m_FromVideo      = audio->m_FromVideo;
		m_FileName       = audio->m_FileName;
		m_BitRate        = audio->m_BitRate;
		m_Opener         = audio->m_Opener;
		m_PreProcessor   = audio->m_PreProcessor;
		m_IsChecked      = audio->m_IsChecked;
		m_Info           = audio->m_Info;
		m_FileSelected   = audio->m_FileSelected;
		m_TrackSelected  = audio->m_TrackSelected;
	} else {
		m_FromVideo      = false;
		m_BitRate        = 0;
		m_Opener         = g_Config->GetAudioOpener();
		m_PreProcessor   = g_Config->GetAudioPreProcessor();
		SetInvalid(true);
		m_FileSelected   = false;
		m_TrackSelected  = false;
	}
	m_Index = -1;
}

Audio::~Audio()
{
}

Video * Audio::GetVideo() const
{
	return m_Video;
}

bool Audio::IsFileSelected() const
{
	return m_FileSelected;
}


void Audio::SetFileSelected(bool selected)
{
	m_FileSelected = selected;
}

bool Audio::IsTrackSelected() const
{
	return m_TrackSelected;
}

void Audio::SetTrackSelected(bool selected)
{
	m_TrackSelected = selected;
}

bool Audio::GetFromVideo() const
{
	return m_FromVideo;
}

bool Audio::SetFromVideo(bool fromVideo, bool setChanged)
{
	if (m_FromVideo != fromVideo) {
		m_FromVideo = fromVideo;
		m_IsChecked = false;
		if (setChanged) SetChanged();
	}
	return true;
}

const String & Audio::GetFileName() const
{
	return m_FileName;
}

bool Audio::SetFileName(const String & fileName, bool setChanged)
{
	String fullPath = fileName.toFullPath();
	if (m_FileName != fullPath) {
		m_FileName = fullPath;
		m_IsChecked = false;
		if (setChanged) SetChanged();
	}
	return true;
}

int Audio::GetBitRate() const
{
	return m_BitRate;
}

bool Audio::SetBitRate(int bitRate, bool setChanged)
{
	if (0 <= bitRate && bitRate < AUDIOBITRATE_MAX) {
		if (m_BitRate != bitRate) {
			m_BitRate = bitRate;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

String Audio::GetOpener() const
{
	return (IsScript() ? S("") : m_Opener);
}

bool Audio::SetOpener(const String & opener, bool setChanged)
{
	if (opener.length()) {
		if (m_Opener != opener) {
			m_Opener = opener;
			SetInvalid(true);
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

const String & Audio::GetPreProcessor() const
{
	return m_PreProcessor;
}

bool Audio::SetPreProcessor(const String & preProcessor, bool setChanged)
{
	if (m_PreProcessor != preProcessor) {
		m_PreProcessor = preProcessor;
		SetInvalid(true);
		if (setChanged) SetChanged();
	}
	return true;
}

void Audio::SetInvalid(bool clearChecked)
{
	if (clearChecked) m_IsChecked = false;
	m_Info     = AVSWrapper::Info();
}

bool Audio::Analyze()
{
	m_IsChecked = true;
	SetInvalid(false);

	if (GetFromVideo()) {
		m_Info = GetVideo()->GetInfo();
	} else if (GetFileName().length()) {
		AVSWrapper avs;
		avs.Open(GetFileName(), GetOpener(), false, GetPreProcessor(), 
			(GetVideo()->GetGroup()->GetProject()->GetDVDFormat() == DVDFORMAT_PAL));
		m_Info = avs.GetInfo();
		m_OpenError = avs.GetError();
	}
	Recalculate();
	return IsValid();
}

bool Audio::IsChecked() const
{
	return m_IsChecked;
}

bool Audio::IsValid() const
{
	return m_Info.hasAudio;
}

bool Audio::IsScript() const
{
	if (m_FromVideo) {
		return GetVideo()->IsScript();
	} else {
		return AVSWrapper::IsScript(m_FileName);
	}
}

const AVSWrapper::Info & Audio::GetInfo() const
{
	return m_Info;
}

int Audio::GetNoOfChannels() const
{
	return m_Info.noOfChannels;
}

int Audio::GetBitRateInKbps() const
{
	int bitRate = (GetBitRate() ? GetBitRate() : GetNoOfChannels());
	switch (bitRate) {
		case 0: return 0;
		case 1: return 96;
		case 2: return 192;
		case 3: return 256;
		case 4: return 384;
		default: return 448;
	}
}

int Audio::GetIndex() const
{
	return m_Index;
}

void Audio::SetIndex(int index)
{
	m_Index = index;
}

String Audio::GetIdent(bool forTrack)
{
	if (m_Index < 0) {
		return S("");
	} else if (forTrack) {
		return S("Audio track ") + String(GetIndex() + 1) +
			S(" of title ") + String(GetVideo()->GetTitle()->GetIndex() + 1) +
			S(" in group ") + String(GetVideo()->GetTitle()->GetGroup()->GetIndex() + 1) +
			S(": ");
	} else {
		return S("Audio file ") + String(GetIndex() + 1) +
			S(" of audio track ") + String(GetIndex() + 1) +
			S(" of title ") + String(GetVideo()->GetTitle()->GetIndex() + 1) +
			S(" in group ") + String(GetVideo()->GetTitle()->GetGroup()->GetIndex() + 1) +
			S(": ");
	}
}

String Audio::Recalculate()
{
	String error;
	if (! IsChecked()) Analyze();
	if (! IsValid()) {
		if (GetFromVideo()) {
			error += (GetIdent(false) + GetVideo()->GetFileName().enclose() + S(" doesn't contain an audio track.\n"));
		} else if (GetFileName().length()) {
			error += (GetIdent(false) + GetFileName().enclose() + S(" doesn't seem to be an audio file."));
			if (m_OpenError.length()) error += S(" (") + m_OpenError + S(")");
			error += S("\n");
		} else {
			error += (GetIdent(false) + S("No audio file is given.\n"));
		}
	}
	return error;
}

void Audio::SetChanged()
{
	m_Video->SetChanged();
}

void Audio::GetDescription(StringVectorVector & strings, bool forTitle) const
{
	StringVector line;
	if (forTitle) {
		if (GetVideo()->GetGroup()->GetView() == VIEW_TITLETRACKPART) {
			line.push_back(S("Audio track ") + String(GetIndex() + 1) + S(": "));
		} else {
			line.push_back(S("Title ") + String(GetVideo()->GetTitle()->GetIndex() + 1) + S(": "));
		}
		String s = (GetVideo()->GetTitle()->GetConvertToMono(GetIndex()) ? S("mono") : GetNoOfChannels()) + S(" channels")
			+ S(", ") + String(GetBitRateInKbps()) + S(" kb/s");
		line.push_back(s);
		strings.push_back(line);
	} else {
		String s = S("Audio file");
		if (GetVideo()->GetTitle()->GetVideoCount() > 1) {
			s += S(" ") + String(GetVideo()->GetIndex() + 1);
		}
		s += S(": ");
		line.push_back(s);
		if (GetFromVideo()) {
			s = S("audio track from video file") + (m_PreProcessor.length() ? S("(+") + m_PreProcessor  + S(")") : S(""));
		} else {
			s = GetFileName().toFileName() + S(" (") + m_Opener + 
				(m_PreProcessor.length() ? S("+") + m_PreProcessor : S("")) + S(")");
		}
		line.push_back(s);
		strings.push_back(line);
		line.clear();
		line.push_back(String());
		if (IsValid()) {
			s = String(GetNoOfChannels()) + S(" channels");
			line.push_back(s);
		} else {
			if (GetFromVideo()) {
				line.push_back(S("Error: the video file doesn't have an audio track."));
			} else {
				line.push_back(S("Error: ") + m_OpenError);
			}
		}
		strings.push_back(line);
	}
}

bool Audio::Load(csv_file_t csv, int & lineno, int nest)
{
	bool ok = true;

	int field = 1;
	ok = ok && SetFromVideo(csv_get_int(csv, lineno, ++field, m_FromVideo) != 0, false);
	ok = ok && SetFileName(String().fromUtf8(csv_get_str(csv, lineno, ++field, m_FileName.toUtf8().c_str())), false);
	field++;
	field++;
	ok = ok && SetBitRate(csv_get_int(csv, lineno, ++field, m_BitRate), false);
	SetFileSelected(csv_get_int(csv, lineno, ++field, m_FileSelected ? 1 : 0) != 0);
	SetTrackSelected(csv_get_int(csv, lineno, ++field, m_TrackSelected ? 1 : 0) != 0);
	ok = ok && SetOpener(String().fromUtf8(csv_get_str(csv, lineno, ++field, m_Opener.toUtf8().c_str())), false);
	ok = ok && SetPreProcessor(String().fromUtf8(csv_get_str(csv, lineno, ++field, m_PreProcessor.toUtf8().c_str())), false);

	lineno++;
	while (ok && lineno < csv->count && csv_get_int(csv, lineno, 0, 0) >= nest) {
		lineno++;
	}

	return ok;
}

void Audio::Save(FILE * f, int nest) const
{
	String fileName = String(GetAllocRelativePath(m_FileName.c_str(), NULL), true);
	fprintf(f, "%d,audio,%d,%s,%d,%s,%d,%d,%d,%s,%s\n", nest,
		(m_FromVideo ? 1 : 0),
		fileName.enclose(TEXT('"')).toUtf8().c_str(),
		0,
		"",
		m_BitRate,
		(m_FileSelected ? 1 : 0),
		(m_TrackSelected ? 1 : 0),
		m_Opener.enclose().toUtf8().c_str(),
		m_PreProcessor.enclose().toUtf8().c_str());
}
