#include "main.h"
#include "config.h"
#include "project.h"
#include "menuitem.h"

#include "safefgets/safefgets.h"
#include "winutils/winutils.h"
#include "csvread/csvread.h"

#include <shlobj.h>
#include <stdio.h>

Config::Config()
{
	m_DVDFormat              = DVDFORMAT_PAL;
	m_DVDSize                = DVDSIZE_DVD_5;
	m_CustomSize             = 4400;
	m_CustomBitRate          = 5000;
	m_OutputType             = OUTPUTTYPE_DVDFOLDER;
	m_HasChanged             = false;
	m_MenuColor[MENUCOLOR_BACKGROUND] = RGB(0, 0, 0);
	m_MenuColor[MENUCOLOR_NORMAL] = RGB(192, 192, 192);
	m_MenuColor[MENUCOLOR_HILITE] = RGB(255, 255, 255);
	m_MenuColor[MENUCOLOR_SELECT] = RGB(255, 255, 128);
	m_PlayTextFromFileName   = false;
	m_SelectTextFromLanguage = false;
	m_NormalText             = true;
	m_HiliteText             = true;
	m_MenuFontFace           = S("Arial");
	m_MenuFontSize           = 32;
	m_MenuBold               = false;
	m_MenuItalic             = false;
	m_NormalRect             = false;
	m_HiliteRect             = false;
	m_Thickness              = 1;
	m_HRadius                = 0;
	m_VRadius                = 0;
	m_FillNormal             = false;
	m_FillHilite             = false;
	m_AutoResize             = true;
	m_HMargin                = 0;
	m_VMargin                = 0;
	m_Alignment              = ALIGN_CENTERCENTER;
	m_CompensateOverscan     = false;
	m_OverscanPercent        = 5;
	m_CreateChapters         = false;
	m_ChapterCount           = 5;
	m_FontFace               = S("Arial");
	m_FontSize               = 32;
	m_Bold                   = false;
	m_Italic                 = false;
	m_Box                    = false;
	m_Color                  = RGB(255, 255, 255);
	m_Encoder                = ENCODER_HC;
	m_HideEncoder            = false;
	m_HCProfile              = HCPROFILE_BEST;
	m_Qu2Pass                = true;
	m_QuHighQuality          = true;
	m_QuTrellis              = false;
	m_QuExtreme              = false;
	m_VideoOpener            = S("DirectShowOpener");
	m_AudioOpener            = S("DirectShowOpener");
	m_ImgBurn = String(GetFullPathInSpecialFolder(CSIDL_PROGRAM_FILES, NULL, 
		TEXT("ImgBurn"), TEXT("ImgBurn.exe"), NULL, 0, FALSE, NULL), true);
	m_HasChanged = false;

	m_ConfigFile = String(GetFullPathInSpecialFolder(CSIDL_APPDATA, NULL, 
		TEXT("DVDforger"), TEXT("config.dfc"), NULL, 0, TRUE, NULL), true);
	Load(m_ConfigFile);

	m_FunctionDir = String(GetFullPathInSpecialFolder(CSIDL_APPDATA, NULL, 
		TEXT("DVDforger"), TEXT("Functions"), NULL, 0, TRUE, NULL), true);
	m_FunctionMask = String(GetFullPathInSpecialFolder(CSIDL_APPDATA, NULL, 
		TEXT("DVDforger\\Functions"), TEXT("*.avsi"), NULL, 0, TRUE, NULL), true);
	String defaultOpener = m_FunctionDir + S("\\DefaultOpener.avsi");
	if (GetFileAttributes(defaultOpener.c_str()) == 0xFFFFFFFF) {
		FILE * f = fopen(defaultOpener.toAnsi().c_str(), "wt");
		fprintf(f, "# This is the default file opener function of DVDforger.\n");
		fprintf(f, "# Use this as an example to write your own opener functions.\n");
		fprintf(f, "# An opener function must receive the file name as a string\n");
		fprintf(f, "# parameter and whether it shall be PAL or NTSC as a bool\n");
		fprintf(f, "# parameter (even if the latter isn't used).\n");
		fprintf(f, "\n");
		fprintf(f, "function DirectShowOpener(string fileName, bool PAL)\n");
		fprintf(f, "{\n");
		fprintf(f, "  try {\n");
		fprintf(f, "    return DirectShowSource(fileName, seek=false)\n");
		fprintf(f, "  } catch (err_msg) {\n");
		fprintf(f, "    return DirectShowSource(fileName, seek=false, convertfps=true, fps=(PAL ? 25.0 : 30000.0 / 1001))\n");
		fprintf(f, "  }\n");
		fprintf(f, "}\n");
		fprintf(f, "\n");
		fprintf(f, "function AviOpener(string fileName, bool PAL)\n");
		fprintf(f, "{\n");
		fprintf(f, "  return EnsureVBRMP3Sync(AVISource(fileName))\n");
		fprintf(f, "}\n");
		fclose(f);
		SetFileAttributes(defaultOpener.c_str(), FILE_ATTRIBUTE_READONLY);
	}
	String examplePreProcessor = m_FunctionDir + S("\\ExamplePreProcessor.avsi");
	if (GetFileAttributes(examplePreProcessor.c_str()) == 0xFFFFFFFF) {
		FILE * f = fopen(examplePreProcessor.toAnsi().c_str(), "wt");
		fprintf(f, "# This is an example pre-processor function.\n");
		fprintf(f, "# Use this as an example to write your own pre-processor functions.\n");
		fprintf(f, "# A pre-processor function must receive a clip as its first\n");
		fprintf(f, "# parameter and whether it shall be PAL or NTSC as a bool\n");
		fprintf(f, "# parameter (even if the latter isn't used).\n");
		fprintf(f, "\n");
		fprintf(f, "function FadeIn10seconds(clip c, bool PAL) {\n");
		fprintf(f, "  return FadeIn0(c, Int(c.FrameRate * 10))\n");
		fprintf(f, "}\n");
		fclose(f);
	}
	RefreshFunctionMaps();
}

Config::~Config()
{
}

bool Config::SetDVDFormat(int dvdFormat, bool setChanged)
{
	if (0 <= dvdFormat && dvdFormat < DVDFORMAT_MAX) {
		if (m_DVDFormat != dvdFormat) {
			m_DVDFormat = dvdFormat;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

int Config::GetDVDFormat() const
{
	return m_DVDFormat;
}

bool Config::SetDVDSize(int dvdSize, bool setChanged)
{
	if (0 <= dvdSize && dvdSize < DVDSIZE_MAX) {
		if (m_DVDSize != dvdSize) {
			m_DVDSize = dvdSize;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

int Config::GetDVDSize() const
{
	return m_DVDSize;
}

bool Config::SetCustomSize(int customSize, bool setChanged)
{
	if (0 < customSize) {
		if (m_CustomSize != customSize) {
			m_CustomSize = customSize;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return true;
}

int Config::GetCustomSize() const
{
	return m_CustomSize;
}

bool Config::SetCustomBitRate(int customBitRate, bool setChanged)
{
	if (0 < customBitRate) {
		if (customBitRate > 9000) customBitRate = 9000;
		if (m_CustomBitRate != customBitRate) {
			m_CustomBitRate = customBitRate;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return true;
}

int Config::GetCustomBitRate() const
{
	return m_CustomBitRate;
}

bool Config::SetOutputType(int outputType, bool setChanged)
{
	if (0 <= outputType && outputType < OUTPUTTYPE_MAX) {
		if (m_OutputType != outputType) {
			m_OutputType = outputType;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

int Config::GetOutputType() const
{
	return m_OutputType;
}

bool Config::SetPlayTextFromFileName(bool playTextFromFileName, bool setChanged)
{
	if (m_PlayTextFromFileName != playTextFromFileName) {
		m_PlayTextFromFileName = playTextFromFileName;
		if (setChanged) SetChanged();
	}
	return true;
}

COLORREF Config::GetMenuColor(int which) const
{
	return m_MenuColor[which];
}

bool Config::SetMenuColor(int which, COLORREF color, bool setChanged)
{
	if (m_MenuColor[which] != color) {
		m_MenuColor[which] = color;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Config::GetPlayTextFromFileName() const
{
	return m_PlayTextFromFileName;
}

bool Config::SetSelectTextFromLanguage(bool selectTextFromLanguage, bool setChanged)
{
	if (m_SelectTextFromLanguage != selectTextFromLanguage) {
		m_SelectTextFromLanguage = selectTextFromLanguage;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Config::GetSelectTextFromLanguage() const
{
	return m_SelectTextFromLanguage;
}

bool Config::GetNormalText() const
{
	return m_NormalText;
}

bool Config::SetNormalText(bool normalText, bool setChanged)
{
	if (m_NormalText != normalText) {
		m_NormalText = normalText;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Config::GetHiliteText() const
{
	return m_HiliteText;
}

bool Config::SetHiliteText(bool hiliteText, bool setChanged)
{
	if (m_HiliteText != hiliteText) {
		m_HiliteText = hiliteText;
		if (setChanged) SetChanged();
	}
	return true;
}

const String & Config::GetMenuFontFace()
{
	return m_MenuFontFace;
}

bool Config::SetMenuFontFace(const String & fontFace, bool setChanged)
{
	if (m_MenuFontFace != fontFace) {
		m_MenuFontFace = fontFace;
		if (setChanged) SetChanged();
	}
	return true;
}

int Config::GetMenuFontSize() const
{
	return m_MenuFontSize;
}

bool Config::SetMenuFontSize(int fontSize, bool setChanged)
{
	if (0 <= fontSize) {
		if (m_MenuFontSize != fontSize) {
			m_MenuFontSize = fontSize;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

bool Config::GetMenuBold() const
{
	return m_MenuBold;
}

bool Config::SetMenuBold(bool bold, bool setChanged)
{
	if (m_MenuBold != bold) {
		m_MenuBold = bold;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Config::GetMenuItalic() const
{
	return m_MenuItalic;
}

bool Config::SetMenuItalic(bool italic, bool setChanged)
{
	if (m_MenuItalic != italic) {
		m_MenuItalic = italic;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Config::GetNormalRect() const
{
	return m_NormalRect;
}

bool Config::SetNormalRect(bool normalRect, bool setChanged)
{
	if (m_NormalRect != normalRect) {
		m_NormalRect = normalRect;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Config::GetHiliteRect() const
{
	return m_HiliteRect;
}

bool Config::SetHiliteRect(bool hiliteRect, bool setChanged)
{
	if (m_HiliteRect != hiliteRect) {
		m_HiliteRect = hiliteRect;
		if (setChanged) SetChanged();
	}
	return true;
}

int Config::GetThickness() const
{
	return m_Thickness;
}

bool Config::SetThickness(int thickness, bool setChanged)
{
	if (0 < thickness) {
		if (m_Thickness != thickness) {
			m_Thickness = thickness;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

int Config::GetHRadius() const
{
	return m_HRadius;
}

bool Config::SetHRadius(int hRadius, bool setChanged)
{
	if (0 <= hRadius) {
		if (m_HRadius != hRadius) {
			m_HRadius = hRadius;
			if (! m_VRadius) m_VRadius = hRadius;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

int Config::GetVRadius() const
{
	return m_VRadius;
}

bool Config::SetVRadius(int vRadius, bool setChanged)
{
	if (0 <= vRadius) {
		if (m_VRadius != vRadius) {
			m_VRadius = vRadius;
			if (! m_HRadius) m_HRadius = vRadius;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

bool Config::GetFillNormal() const
{
	return m_FillNormal;
}

bool Config::SetFillNormal(bool fillNormal, bool setChanged)
{
	if (m_FillNormal != fillNormal) {
		m_FillNormal = fillNormal;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Config::GetFillHilite() const
{
	return m_FillHilite;
}

bool Config::SetFillHilite(bool fillHilite, bool setChanged)
{
	if (m_FillHilite != fillHilite) {
		m_FillHilite = fillHilite;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Config::GetAutoResize() const
{
	return m_AutoResize;
}

bool Config::SetAutoResize(bool autoResize, bool setChanged)
{
	if (m_AutoResize != autoResize) {
		m_AutoResize = autoResize;
		if (setChanged) SetChanged();
	}
	return true;
}

int Config::GetHMargin() const
{
	return m_HMargin;
}

bool Config::SetHMargin(int hMargin, bool setChanged)
{
	if (0 <= hMargin) {
		if (m_HMargin != hMargin) {
			m_HMargin = hMargin;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

int Config::GetVMargin() const
{
	return m_VMargin;
}

bool Config::SetVMargin(int vMargin, bool setChanged)
{
	if (0 <= vMargin) {
		if (m_VMargin != vMargin) {
			m_VMargin = vMargin;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

int Config::GetAlignment() const
{
	return m_Alignment;
}

bool Config::SetAlignment(int alignment, bool setChanged)
{
	if (0 <= alignment && alignment < ALIGN_MAX) {
		if (m_Alignment != alignment) {
			m_Alignment = alignment;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

bool Config::GetCompensateOverscan() const
{
	return m_CompensateOverscan;
}

bool Config::SetCompensateOverscan(bool compensateOverscan, bool setChanged)
{
	if (m_CompensateOverscan != compensateOverscan) {
		m_CompensateOverscan = compensateOverscan;
		if (setChanged) SetChanged();
	}
	return true;
}
		
int Config::GetOverscanPercent() const
{
	return m_OverscanPercent;
}

bool Config::SetOverscanPercent(int overscanPercent, bool setChanged)
{
	if (0 <= overscanPercent && overscanPercent < 100) {
		if (m_OverscanPercent != overscanPercent) {
			m_OverscanPercent = overscanPercent;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

bool Config::GetCreateChapters() const
{
	return m_CreateChapters;
}

bool Config::SetCreateChapters(bool createChapters, bool setChanged)
{
	if (m_CreateChapters != createChapters) {
		m_CreateChapters = createChapters;
		if (setChanged) SetChanged();
	}
	return true;
}

int  Config::GetChapterCount() const
{
	return m_ChapterCount;
}

bool Config::SetChapterCount(int chapterCount, bool setChanged)
{
	if (0 <= chapterCount) {
		if (m_ChapterCount != chapterCount) {
			m_ChapterCount = chapterCount;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

const String & Config::GetFontFace() const
{
	return m_FontFace;
}

bool Config::SetFontFace(const String & fontFace, bool setChanged)
{
	if (m_FontFace != fontFace) {
		m_FontFace = fontFace;
		if (setChanged) SetChanged();
	}
	return true;
}

int Config::GetFontSize() const
{
	return m_FontSize;
}

bool Config::SetFontSize(int fontSize, bool setChanged)
{
	if (0 < fontSize) {
		if (m_FontSize != fontSize) {
			m_FontSize = fontSize;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

bool Config::GetBold() const
{
	return m_Bold;
}

bool Config::SetBold(bool bold, bool setChanged)
{
	if (m_Bold != bold) {
		m_Bold = bold;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Config::GetItalic() const
{
	return m_Italic;
}

bool Config::SetItalic(bool italic, bool setChanged)
{
	if (m_Italic != italic) {
		m_Italic = italic;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Config::GetBox() const
{
	return m_Box;
}

bool Config::SetBox(bool box, bool setChanged)
{
	if (m_Box != box) {
		m_Box = box;
		if (setChanged) SetChanged();
	}
	return true;
}

COLORREF Config::GetColor() const
{
	return m_Color;
}

bool Config::SetColor(COLORREF color, bool setChanged)
{
	if (m_Color != color) {
		m_Color = color;
		if (setChanged) SetChanged();
	}
	return true;
}

int Config::GetEncoder() const
{
	return m_Encoder;
}

bool Config::SetEncoder(int encoder, bool setChanged)
{
	if (0 <= encoder && encoder < ENCODER_MAX) {
		if (m_Encoder != encoder) {
			m_Encoder = encoder;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

bool Config::GetHideEncoder() const
{
	return m_HideEncoder;
}

bool Config::SetHideEncoder(bool hideEncoder, bool setChanged)
{
	if (m_HideEncoder != hideEncoder) {
		m_HideEncoder = hideEncoder;
		if (setChanged) SetChanged();
	}
	return true;
}

int Config::GetHCProfile() const
{
	return m_HCProfile;
}

bool Config::SetHCProfile(int hcProfile, bool setChanged)
{
	if (0 <= hcProfile && hcProfile < HCPROFILE_MAX) {
		if (m_HCProfile != hcProfile) {
			m_HCProfile = hcProfile;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

const String & Config::GetHCOptions() const
{
	return m_HCOptions;
}

bool Config::SetHCOptions(const String & hcOptions, bool setChanged)
{
	if (m_HCOptions != hcOptions) {
		m_HCOptions = hcOptions;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Config::GetQu2Pass() const
{
	return m_Qu2Pass;
}

bool Config::SetQu2Pass(bool qu2Pass, bool setChanged)
{
	if (m_Qu2Pass != qu2Pass) {
		m_Qu2Pass = qu2Pass;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Config::GetQuHighQuality() const
{
	return m_QuHighQuality;
}

bool Config::SetQuHighQuality(bool quHighQuality, bool setChanged)
{
	if (m_QuHighQuality != quHighQuality) {
		m_QuHighQuality = quHighQuality;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Config::GetQuTrellis() const
{
	return m_QuTrellis;
}

bool Config::SetQuTrellis(bool quTrellis, bool setChanged)
{
	if (m_QuTrellis != quTrellis) {
		m_QuTrellis = quTrellis;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Config::GetQuExtreme() const
{
	return m_QuExtreme;
}

bool Config::SetQuExtreme(bool quExtreme, bool setChanged)
{
	if (m_QuExtreme != quExtreme) {
		m_QuExtreme = quExtreme;
		if (setChanged) SetChanged();
	}
	return true;
}

const String & Config::GetQuParams() const
{
	return m_QuParams;
}

bool Config::SetQuParams(const String & quParams, bool setChanged)
{
	if (m_QuParams != quParams) {
		m_QuParams = quParams;
		if (setChanged) SetChanged();
	}
	return true;
}

const String & Config::GetAftenParams() const
{
	return m_AftenParams;
}

bool Config::SetAftenParams(const String & aftenParams, bool setChanged)
{
	if (m_AftenParams != aftenParams) {
		m_AftenParams = aftenParams;
		if (setChanged) SetChanged();
	}
	return true;
}

const String & Config::GetVideoOpener() const
{
	return m_VideoOpener;
}

bool Config::SetVideoOpener(const String & opener, bool setChanged)
{
	if (opener.length()) {
		if (m_VideoOpener != opener) {
			m_VideoOpener = opener;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

const String & Config::GetVideoPreProcessor() const
{
	return m_VideoPreProcessor;
}

bool Config::SetVideoPreProcessor(const String & preProcessor, bool setChanged)
{
	if (m_VideoPreProcessor != preProcessor) {
		m_VideoPreProcessor = preProcessor;
		if (setChanged) SetChanged();
	}
	return true;
}

const String & Config::GetImgBurn() const
{
	return m_ImgBurn;
}

bool Config::SetImgBurn(const String & imgBurn, bool setChanged)
{
	String fullPath = imgBurn.toFullPath();
	if (m_ImgBurn != fullPath) {
		m_ImgBurn = fullPath;
		if (setChanged) SetChanged();
	}
	return true;
}

const String & Config::GetAudioOpener() const
{
	return m_AudioOpener;
}

bool Config::SetAudioOpener(const String & opener, bool setChanged)
{
	if (opener.length()) {
		if (m_AudioOpener != opener) {
			m_AudioOpener = opener;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

const String & Config::GetAudioPreProcessor() const
{
	return m_AudioPreProcessor;
}

bool Config::SetAudioPreProcessor(const String & preProcessor, bool setChanged)
{
	if (m_AudioPreProcessor != preProcessor) {
		m_AudioPreProcessor = preProcessor;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Config::IsImgBurnValid() const
{
	DWORD attributes = GetFileAttributes(m_ImgBurn.c_str());
	return ((attributes != 0xFFFFFFFF) && ! (attributes & FILE_ATTRIBUTE_DIRECTORY));
}

bool Config::HasChanged() const
{
	return m_HasChanged;
}

void Config::SetChanged()
{
	m_HasChanged = true;
}

void Config::ClearChanged()
{
	m_HasChanged = false;
}

bool Config::Load(const String & fileName)
{
	FILE * f = fopen(fileName.toAnsi().c_str(), "rt");
	if (! f) return false;
	csv_file_t csv = csv_read(f, ',', '"', '"');
	fclose(f);
	bool ok = ! strcmp(csv_get_str(csv, 0, 0, ""), "DVDforgerConfig");
	int field = 0;
	ok = ok && SetDVDFormat(csv_get_int(csv, 0, ++field, m_DVDFormat), false);
	ok = ok && SetDVDSize(csv_get_int(csv, 0, ++field, m_DVDSize), false);
	ok = ok && SetCustomSize(csv_get_int(csv, 0, ++field, m_CustomSize), false);
	ok = ok && SetCustomBitRate(csv_get_int(csv, 0, ++field, m_CustomBitRate), false);
	ok = ok && SetOutputType(csv_get_int(csv, 0, ++field, m_OutputType), false);
	ok = ok && SetPlayTextFromFileName(csv_get_int(csv, 0, ++field, m_PlayTextFromFileName ? 1 : 0) != 0, false);
	ok = ok && SetSelectTextFromLanguage(csv_get_int(csv, 0, ++field, m_SelectTextFromLanguage ? 1 : 0) != 0, false);
	ok = ok && SetNormalText(csv_get_int(csv, 0, ++field, m_NormalText ? 1 : 0) != 0, false);
	ok = ok && SetHiliteText(csv_get_int(csv, 0, ++field, m_HiliteText ? 1 : 0) != 0, false);
	for (int i = 0; i < 4; i++) ok = ok && SetMenuColor(i, csv_get_int(csv, 0, ++field, m_MenuColor[i]), false);
	ok = ok && SetMenuFontFace(String().fromUtf8(csv_get_str(csv, 0, ++field, m_MenuFontFace.toUtf8().c_str())), false);
	ok = ok && SetMenuFontSize(csv_get_int(csv, 0, ++field, m_MenuFontSize), false);
	ok = ok && SetMenuBold(csv_get_int(csv, 0, ++field, m_MenuBold ? 1 : 0) != 0, false);
	ok = ok && SetMenuItalic(csv_get_int(csv, 0, ++field, m_MenuItalic ? 1 : 0) != 0, false);
	ok = ok && SetNormalRect(csv_get_int(csv, 0, ++field, m_NormalRect ? 1 : 0) != 0, false);
	ok = ok && SetHiliteRect(csv_get_int(csv, 0, ++field, m_HiliteRect ? 1 : 0) != 0, false);
	ok = ok && SetThickness(csv_get_int(csv, 0, ++field, m_Thickness), false);
	ok = ok && SetHRadius(csv_get_int(csv, 0, ++field, m_HRadius), false);
	ok = ok && SetVRadius(csv_get_int(csv, 0, ++field, m_VRadius), false);
	ok = ok && SetFillNormal(csv_get_int(csv, 0, ++field, m_FillNormal ? 1 : 0) != 0, false);
	ok = ok && SetFillHilite(csv_get_int(csv, 0, ++field, m_FillHilite ? 1 : 0) != 0, false);
	ok = ok && SetAutoResize(csv_get_int(csv, 0, ++field, m_AutoResize ? 1 : 0) != 0, false);
	ok = ok && SetHMargin(csv_get_int(csv, 0, ++field, m_HMargin), false);
	ok = ok && SetVMargin(csv_get_int(csv, 0, ++field, m_VMargin), false);
	ok = ok && SetAlignment(csv_get_int(csv, 0, ++field, m_Alignment), false);
	ok = ok && SetCompensateOverscan(csv_get_int(csv, 0, ++field, m_CompensateOverscan ? 1 : 0) != 0, false);
	ok = ok && SetOverscanPercent(csv_get_int(csv, 0, ++field, m_OverscanPercent), false);
	ok = ok && SetCreateChapters(csv_get_int(csv, 0, ++field, m_CreateChapters ? 1 : 0) != 0, false);
	ok = ok && SetChapterCount(csv_get_int(csv, 0, ++field, m_ChapterCount), false);
	ok = ok && SetFontFace(String().fromUtf8(csv_get_str(csv, 0, ++field, m_FontFace.toUtf8().c_str())), false);
	ok = ok && SetFontSize(csv_get_int(csv, 0, ++field, m_FontSize), false);
	ok = ok && SetBold(csv_get_int(csv, 0, ++field, m_Bold ? 1 : 0) != 0, false);
	ok = ok && SetItalic(csv_get_int(csv, 0, ++field, m_Italic ? 1 : 0) != 0, false);
	ok = ok && SetBox(csv_get_int(csv, 0, ++field, m_Box ? 1 : 0) != 0, false);
	ok = ok && SetColor(csv_get_int(csv, 0, ++field, m_Color), false);
	ok = ok && SetEncoder(csv_get_int(csv, 0, ++field, m_Encoder), false);
	ok = ok && SetHideEncoder(csv_get_int(csv, 0, ++field, m_HideEncoder ? 1 : 0) != 0, false);
	ok = ok && SetHCProfile(csv_get_int(csv, 0, ++field, m_HCProfile), false);
	ok = ok && SetHCOptions(String().fromUtf8(csv_get_str(csv, 0, ++field, m_HCOptions.toUtf8().c_str())), false);
	ok = ok && SetQu2Pass(csv_get_int(csv, 0, ++field, m_Qu2Pass ? 1 : 0) != 0, false);
	ok = ok && SetQuHighQuality(csv_get_int(csv, 0, ++field, m_QuHighQuality ? 1 : 0) != 0, false);
	ok = ok && SetQuTrellis(csv_get_int(csv, 0, ++field, m_QuTrellis ? 1 : 0) != 0, false);
	ok = ok && SetQuExtreme(csv_get_int(csv, 0, ++field, m_QuExtreme ? 1 : 0) != 0, false);
	ok = ok && SetQuParams(String().fromUtf8(csv_get_str(csv, 0, ++field, m_QuParams.toUtf8().c_str())), false);
	ok = ok && SetAftenParams(String().fromUtf8(csv_get_str(csv, 0, ++field, m_AftenParams.toUtf8().c_str())), false);
	ok = ok && SetVideoOpener(String().fromUtf8(csv_get_str(csv, 0, ++field, m_VideoOpener.toUtf8().c_str())), false);
	ok = ok && SetVideoPreProcessor(String().fromUtf8(csv_get_str(csv, 0, ++field, m_VideoPreProcessor.toUtf8().c_str())), false);
	ok = ok && SetAudioOpener(String().fromUtf8(csv_get_str(csv, 0, ++field, m_AudioOpener.toUtf8().c_str())), false);
	ok = ok && SetAudioPreProcessor(String().fromUtf8(csv_get_str(csv, 0, ++field, m_AudioPreProcessor.toUtf8().c_str())), false);
	ok = ok && SetImgBurn(String().fromUtf8(csv_get_str(csv, 0, ++field, m_ImgBurn.toUtf8().c_str())), false);
	csv_free(csv);
	return ok;
}

bool Config::Save(String fileName)
{
	bool ok = false;
	if (! fileName.length()) fileName = m_ConfigFile;
	FILE * f = fopen(fileName.toAnsi().c_str(), "wt");
	if (f) {
		fprintf(f, "DVDforgerConfig,%d,%d,%d,%d,%d",
			m_DVDFormat, m_DVDSize, m_CustomSize, m_CustomBitRate, m_OutputType);
		fprintf(f, ",%d,%d,%d,%d,%d,%d,%d,%d",
			m_PlayTextFromFileName ? 1 : 0, m_SelectTextFromLanguage ? 1 : 0,
			m_NormalText ? 1 : 0, m_HiliteText ? 1 : 0,
			m_MenuColor[0], m_MenuColor[1], m_MenuColor[2], m_MenuColor[3]);
		fprintf(f, ",%s,%d,%d,%d,%d,%d",
			m_MenuFontFace.enclose().toUtf8().c_str(), m_MenuFontSize,
			m_MenuBold ? 1 : 0, m_MenuItalic ? 1 : 0, m_NormalRect ? 1 : 0, m_HiliteRect ? 1 : 0);
		fprintf(f, ",%d,%d,%d,%d,%d,%d,%d,%d,%d",
			m_Thickness, m_HRadius, m_VRadius, m_FillNormal ? 1 : 0, m_FillHilite ? 1 : 0,
			m_AutoResize ? 1 : 0, m_HMargin, m_VMargin, m_Alignment);
		fprintf(f, ",%d,%d,%d,%d",
			m_CompensateOverscan ? 1 : 0, m_OverscanPercent, m_CreateChapters ? 1 : 0, m_ChapterCount);
		fprintf(f, ",%s,%d,%d,%d,%d,%d",
			m_FontFace.enclose().toUtf8().c_str(), m_FontSize,
			m_Bold ? 1 : 0, m_Italic ? 1 : 0, m_Box ? 1 : 0, m_Color);
		fprintf(f, ",%d,%d,%d,%s",
			m_Encoder, m_HideEncoder ? 1 : 0, m_HCProfile, 
			m_HCOptions.enclose().toUtf8().c_str());
		fprintf(f, ",%d,%d,%d,%d,%s,%s",
			m_Qu2Pass ? 1 : 0, m_QuHighQuality ? 1 : 0, m_QuTrellis ? 1 : 0, m_QuExtreme ? 1 : 0,
			m_QuParams.enclose().toUtf8().c_str(), m_AftenParams.enclose().toUtf8().c_str());
		fprintf(f, ",%s,%s,%s,%s,%s",
			m_VideoOpener.enclose().toUtf8().c_str(),
			m_VideoPreProcessor.enclose().toUtf8().c_str(),
			m_AudioOpener.enclose().toUtf8().c_str(),
			m_AudioPreProcessor.enclose().toUtf8().c_str(),
			m_ImgBurn.enclose().toUtf8().c_str());
		fprintf(f, "\n");
		ok = ! ferror(f);
		if (fclose(f)) ok = false;
	}
	return ok;
}

const String & Config::GetFunctionDir()
{
	return m_FunctionDir;
}

bool Config::RefreshFunctionMaps()
{
	unsigned hashValue = 0;
	int shift = -1;
	m_Openers.clear();
	m_PreProcessors.clear();
	WIN32_FIND_DATA findData;
	HANDLE findHandle = FindFirstFile(m_FunctionMask.c_str(), & findData);
	if (findHandle != INVALID_HANDLE_VALUE) {
		do {
			String fileName = String(m_FunctionDir + S("\\") + findData.cFileName).toFullPath();
			FILE * f = fopen(fileName.toAnsi().c_str(), "rt");
			char * line = NULL;
			while (line = safefgets(line, f)) {
				StringVector token = String().fromAnsi(line).tokenize();
				if (token.size() > 8
					&& ! lstrcmpi(token[0].c_str(), TEXT("function"))
					&& ! lstrcmpi(token[2].c_str(), TEXT("("))
					&& ! lstrcmpi(token[5].c_str(), TEXT(","))
					&& ! lstrcmpi(token[6].c_str(), TEXT("bool"))
					&& ! lstrcmpi(token[8].c_str(), TEXT(")"))) {
					if (! lstrcmpi(token[3].c_str(), TEXT("string"))) {
						m_Openers[token[1]] = fileName;
					} else if (! lstrcmpi(token[3].c_str(), TEXT("clip"))) {
						m_PreProcessors[token[1]] = fileName;
					}
				}
				for (int i = strlen(line) - 1; i >= 0; i--) {
					hashValue = (hashValue ^ (line[i] << ((shift + 1) % 24)));
				}
			}
			fclose(f);
		} while (FindNextFile(findHandle, & findData));
	}
	bool changed = (m_HashValue != hashValue);
	m_HashValue = hashValue;
	return changed;
}

const Config::FunctionMap & Config::GetOpeners()
{
	return m_Openers;
}

const Config::FunctionMap & Config::GetPreProcessors()
{
	return m_PreProcessors;
}

