#include "dvdauthor.h"
#include "group.h"
#include "title.h"
#include "video.h"
#include "audio.h"
#include "subtitle.h"
#include "menu.h"
#include "menuitem.h"
#include "tools.h"
#include "langcode.h"

#include "safefgets/safefgets.h"
#include "winutils/winutils.h"

#include <vector>
#include <algorithm>

DVDAuthor::DVDAuthor(ProjectPtr project) : Thread()
{
	m_Project = project;
}

DVDAuthor::~DVDAuthor()
{
}

String DVDAuthor::GetDescription() const
{
	return S("Authoring DVD");
}

String DVDAuthor::CreateChapters(int gindex, int tindex)
{
	String chapters = String().fromLength(0.0);

	String scdFileName = String(m_Project->GetOutputDirectory() + S("\\set") + String(gindex + 1) 
		+ S("title") + String(tindex + 1) + S("video_scd.txt")).toFullPath();
	AddFileToDelete(scdFileName);

	TitlePtr title = m_Project->GetGroup(gindex)->GetTitle(tindex);

	std::vector<int> sceneChanges;
	char * line = NULL;
	FILE * f = fopen(scdFileName.toAnsi().c_str(), "rt");
	if (f) {
		while (line = safefgets(line, f)) {
			int frame;
			if (sscanf(line, "%d", & frame) == 1) {
				sceneChanges.push_back(frame);
			}
		}
		fclose(f);
	}
	std::sort(sceneChanges.begin(), sceneChanges.end());

	double frameRate = m_Project->GetFrameRate();
	int noOfFrames   = (int) (title->GetLength() * frameRate);
	std::vector<int> chaptersAt;
	chaptersAt.push_back(0);

	double offset = 0;
	for (int vindex = 0; vindex < title->GetVideoCount(); vindex++) {
		Video * video = title->GetVideo(vindex);
		if (video->GetFixedChapters()) {
			const std::vector<double> & chapterVector = video->GetChapterVector();
			for (int i = 0; i < (int) chapterVector.size(); i++) {
				double at = chapterVector[i];
				if (at < video->GetOrigLength()) {
					at *= video->GetAcceleration();
					at += offset;
					int frame = (int) (at * frameRate);
					if (frame && frame < noOfFrames) {
						chaptersAt.push_back(frame);
					}
				}
			}
		}
		offset += video->GetLength();
	}
	chaptersAt.push_back(noOfFrames);
	std::sort(chaptersAt.begin(), chaptersAt.end());
		
	if (title->GetVideo(0)->GetCreateChapters()) {
		int framesInChapter = (int) (title->GetVideo(0)->GetChapterCount() * 60.0 * frameRate);
		if (framesInChapter) {
			int lastChapter = 0;
			for (std::vector<int>::iterator ci = chaptersAt.begin(); ci != chaptersAt.end(); ci++) {
				int distance = (*ci) - lastChapter;
				if (distance > framesInChapter) {
					int count = distance / framesInChapter + 1;
					int step  = distance / count;
					while (--count) {
						ci = chaptersAt.insert(ci, *(ci - 1) + step);
						ci++;
					}
				}
				lastChapter = (*ci);
			}
		}
	}
		
	int lastFrame = 0;
	for (std::vector<int>::iterator ci = chaptersAt.begin(); ci != chaptersAt.end(); ci++) {
		if (*ci) {
			std::vector<int>::iterator scAfter = std::upper_bound(sceneChanges.begin(), sceneChanges.end(), (*ci));
			if (scAfter != sceneChanges.end()) {
				int nextFrame = (*scAfter);
				if (scAfter != sceneChanges.begin()) {
					std::vector<int>::iterator scBefore = scAfter - 1;
					if ((*ci) - (*scBefore) < (*scAfter) - (*ci)) {
						nextFrame = (*scBefore);
					}
				}
				if (nextFrame > lastFrame) {
					chapters += (S(",") + String().fromLength(nextFrame / frameRate));
					lastFrame = nextFrame;
				}
			}
		}
	}

	return chapters;
}

void DVDAuthor::SwitchAudioSubtitle(FILE * f, int actGroup)
{
	for (int gindex = 0; gindex < m_Project->GetGroupCount(); gindex++) {
		if (gindex != actGroup) {
			fprintf(f, "        if (g12 %% 256 == %d) {\n", gindex);
			fprintf(f, "          g%d = audio %% 256 + subtitle * 256;\n", gindex * 2 + 1);
			fprintf(f, "        }\n");
		}
	}
	fprintf(f, "        if (g12 %% 256 != %d) {\n", actGroup);
	fprintf(f, "          audio = g%d %% 256;\n", actGroup * 2 + 1);
	fprintf(f, "          subtitle = g%d / 256;\n", actGroup * 2 + 1);
	fprintf(f, "          g12 = %d;\n", actGroup);
	fprintf(f, "        }\n");
}

bool DVDAuthor::OnRun()
{
	bool ok = true;

	int menuCount = 0;
	int titleCount = 0;
	for (int gindex = 0; gindex < m_Project->GetGroupCount(); gindex++) {
		int itemCount = 0;
		Group * group = m_Project->GetGroup(gindex);
		if (group->HasMenu()) {
			group->GetMenu()->authorIndex = ++menuCount;
			++itemCount;
			for (int mindex = 0; mindex < group->GetMenu()->GetMenuItemCount(); mindex++) {
				group->GetMenu()->GetMenuItem(mindex)->authorIndex = ++itemCount;
			}
			if (group->GetMenuType() == MENUTYPE_COMBINED) {
				group->GetCombinedMenu()->authorIndex = ++menuCount;
				++itemCount;
				for (int mindex = 0; mindex < group->GetCombinedMenu()->GetMenuItemCount(); mindex++) {
					group->GetCombinedMenu()->GetMenuItem(mindex)->authorIndex = ++itemCount;
				}
			} else if (group->GetMenuType() == MENUTYPE_SEPARATE) {
				if (group->HasAudioMenu()) {
					group->GetAudioMenu()->authorIndex = ++menuCount;
					++itemCount;
					for (int mindex = 0; mindex < group->GetAudioMenu()->GetMenuItemCount(); mindex++) {
						group->GetAudioMenu()->GetMenuItem(mindex)->authorIndex = ++itemCount;
					}
				} 
				if (group->HasSubtitleMenu()) {
					group->GetSubtitleMenu()->authorIndex = ++menuCount;
					++itemCount;
					for (int mindex = 0; mindex < group->GetSubtitleMenu()->GetMenuItemCount(); mindex++) {
						group->GetSubtitleMenu()->GetMenuItem(mindex)->authorIndex = ++itemCount;
					}
				}
			}
		}
		for (int tindex = 0; tindex < group->GetTitleCount(); tindex++) {
			group->GetTitle(tindex)->authorIndex = ++titleCount;
		}
	}
	
	String xmlFileName = String(m_Project->GetOutputDirectory() + S("\\dvdauthor.xml")).toFullPath();
	AddFileToDelete(xmlFileName);

	FILE * f = fopen(xmlFileName.toAnsi().c_str(), "wt");
	if (f) {
		fprintf(f, "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n");
		fprintf(f, "<dvdauthor>\n\n");

		if (m_Project->GetGroup(0)->HasMenu()) {
			fprintf(f, "<vmgm>\n");
			fprintf(f, "  <fpc> {\n");
			for (int gindex = 0; gindex < m_Project->GetGroupCount(); gindex++) {
				GroupPtr group = m_Project->GetGroup(gindex);
				int audioMenuPos = 1;
				int subtitleMenuPos = 1;
				int subOn = (group->GetFirstSubtitleOn() ? 1 : 0);
				if (group->HasMenu()) {
					if (group->GetMenuType() == MENUTYPE_COMBINED) {
						for (int mindex = 0; mindex < group->GetCombinedMenu()->GetMenuItemCount(); mindex++) {
							if (group->GetCombinedMenu()->GetMenuItem(mindex)->GetSelector1() == 0 &&
								group->GetCombinedMenu()->GetMenuItem(mindex)->GetSelector2() == subOn) {
								audioMenuPos = mindex + 1;
							}
						}
					} else if (group->GetMenuType() == MENUTYPE_SEPARATE) {
						for (int mindex = 0; mindex < group->GetAudioMenu()->GetMenuItemCount(); mindex++) {
							if (group->GetAudioMenu()->GetMenuItem(mindex)->GetSelector1() == 0) {
								audioMenuPos = mindex + 1;
							}
						}
						for (int mindex = 0; mindex < group->GetSubtitleMenu()->GetMenuItemCount(); mindex++) {
							if (group->GetSubtitleMenu()->GetMenuItem(mindex)->GetSelector2() == subOn) {
								subtitleMenuPos = mindex + 1;
							}
						}
					}
				}
				fprintf(f, "    g%d = %d;\n", gindex * 2, 1 + audioMenuPos * 32 + subtitleMenuPos * 1024);
				fprintf(f, "    g%d = %d;\n", gindex * 2 + 1, subOn * 64 * 256);
			}
			fprintf(f, "    g10 = 1;\n");
			fprintf(f, "    g11 = %d;\n", (m_Project->GetGroup(0)->HasMenu() ? 1 : 0));
			fprintf(f, "    g12 = 255;\n");
			fprintf(f, "    jump menu entry title;\n");
			fprintf(f, "  } </fpc>\n");
			fprintf(f, "  <menus>\n");
			fprintf(f, "    <video aspect=\"4:3\"/>\n");
			fprintf(f, "    <pgc entry=\"title\" pause=\"0\">\n");
			fprintf(f, "      <pre> {\n");
			int tsOfMenu = titleCount;
			for (int gindex = 0; gindex < m_Project->GetGroupCount(); gindex++) {
				Group * group = m_Project->GetGroup(gindex);
				if (group->HasMenu()) {
					tsOfMenu++;
					fprintf(f, "        if (g11 == %d) { jump titleset %d menu entry root; }\n", 
						group->GetMenu()->authorIndex, tsOfMenu);
					if (group->GetMenuType() == MENUTYPE_COMBINED) {
						fprintf(f, "        if (g11 == %d) { jump titleset %d menu entry audio; }\n", 
							group->GetCombinedMenu()->authorIndex, tsOfMenu);
					} else if (group->GetMenuType() == MENUTYPE_SEPARATE) {
						if (group->HasAudioMenu()) {
							fprintf(f, "        if (g11 == %d) { jump titleset %d menu entry audio; }\n", 
								group->GetAudioMenu()->authorIndex, tsOfMenu);
						} 
						if (group->HasSubtitleMenu()) {
							fprintf(f, "        if (g11 == %d) { jump titleset %d menu entry subtitle; }\n", 
								group->GetSubtitleMenu()->authorIndex, tsOfMenu);
						}
					}
				}
			}
			for (int gindex = 0; gindex < m_Project->GetGroupCount(); gindex++) {
				Group * group = m_Project->GetGroup(gindex);
				for (int tindex = 0; tindex < group->GetTitleCount(); tindex++) {
					Title * title = group->GetTitle(tindex);
					fprintf(f, "        if (g10 == %d) { jump title %d; }\n", 
						title->authorIndex, title->authorIndex);
				}
			}
			fprintf(f, "      } </pre>\n");
			fprintf(f, "      <vob file=\"%s\" pause=\"0\"/>\n", tools.blank[m_Project->GetDVDFormat()].xmlenc().toUtf8().c_str());
			fprintf(f, "    </pgc>\n");
			fprintf(f, "  </menus>\n");
			fprintf(f, "</vmgm>\n\n");
		} else {
			fprintf(f, "<vmgm/>\n\n");
		}

		for (int gindex = 0; gindex < m_Project->GetGroupCount(); gindex++) {
			GroupPtr group = m_Project->GetGroup(gindex);
			if (group->GetTitleCount()) {
				for (int tindex = 0; tindex < group->GetTitleCount(); tindex++) {
					fprintf(f, "<titleset>\n");
					if (group->HasMenu()) {
						fprintf(f, "  <menus>\n");
						fprintf(f, "    <video aspect=\"%s\"/>\n", group->GetTitle(tindex)->GetVideo(0)->GetFinalAspectRatioStr().toUtf8().c_str());
						fprintf(f, "    <pgc pause=\"0\">\n");
						fprintf(f, "      <pre> { jump vmgm menu entry title; } </pre>\n");
						fprintf(f, "      <vob file=\"%s\" pause=\"0\"/>\n", tools.blank[m_Project->GetDVDFormat()].xmlenc().toUtf8().c_str());
						fprintf(f, "    </pgc>\n");
						fprintf(f, "    <pgc entry=\"root\">\n");
						fprintf(f, "      <pre> {\n");
						fprintf(f, "        g11 = %d;\n", group->GetMenu()->authorIndex);
						fprintf(f, "        jump menu 1;\n");
						fprintf(f, "      } </pre>\n");
						fprintf(f, "    </pgc>\n");
						if (group->GetMenuType() == MENUTYPE_COMBINED) {
							fprintf(f, "    <pgc entry=\"audio\">\n");
							fprintf(f, "      <pre> {\n");
							fprintf(f, "        g11 = %d;\n", group->GetCombinedMenu()->authorIndex);
							fprintf(f, "        jump menu 1;\n");
							fprintf(f, "      } </pre>\n");
							fprintf(f, "    </pgc>\n");
							fprintf(f, "    <pgc entry=\"subtitle\">\n");
							fprintf(f, "      <pre> {\n");
							fprintf(f, "        g11 = %d;\n", group->GetCombinedMenu()->authorIndex);
							fprintf(f, "        jump menu 1;\n");
							fprintf(f, "      } </pre>\n");
							fprintf(f, "    </pgc>\n");
						} else if (group->GetMenuType() == MENUTYPE_SEPARATE) {
							if (group->HasAudioMenu()) {
								fprintf(f, "    <pgc entry=\"audio\">\n");
								fprintf(f, "      <pre> {\n");
								fprintf(f, "        g11 = %d;\n", group->GetAudioMenu()->authorIndex);
								fprintf(f, "        jump menu 1;\n");
								fprintf(f, "      } </pre>\n");
								fprintf(f, "    </pgc>\n");
							} 
							if (group->HasSubtitleMenu()) {
								fprintf(f, "    <pgc entry=\"subtitle\">\n");
								fprintf(f, "      <pre> {\n");
								fprintf(f, "        g11 = %d;\n", group->GetSubtitleMenu()->authorIndex);
								fprintf(f, "        jump menu 1;\n");
								fprintf(f, "      } </pre>\n");
								fprintf(f, "    </pgc>\n");
							}
						}
						fprintf(f, "  </menus>\n");
					}
					fprintf(f, "  <titles>\n");
					fprintf(f, "    <video format=\"%s\" aspect=\"%s\"/>\n",
						(m_Project->GetDVDFormat() == DVDFORMAT_PAL ? "pal" : "ntsc"),
						group->GetTitle(tindex)->GetVideo(0)->GetFinalAspectRatioStr().toUtf8().c_str());
					for (int aindex = 0; aindex < group->GetAudioCount(); aindex++) {
						if (group->GetAudioLanguage(aindex)) {
							fprintf(f, "    <audio lang=\"%s\"/>\n", 
								String(LanguageCode[group->GetAudioLanguage(aindex)]).xmlenc().toUtf8().c_str());
						} else {
							fprintf(f, "    <audio/>\n");
						}
					}
					for (int sindex = 0; sindex < group->GetAudioCount(); sindex++) {
						if (group->GetSubtitleLanguage(sindex)) {
							fprintf(f, "    <subpicture lang=\"%s\"/>\n", 
								String(LanguageCode[group->GetSubtitleLanguage(sindex)]).xmlenc().toUtf8().c_str());
						} else {
							fprintf(f, "    <subpicture/>\n");
						}
					}
					fprintf(f, "    <pgc>\n");
					if (group->HasMenu()) {
						fprintf(f, "      <pre> {\n");
						SwitchAudioSubtitle(f, gindex);
						fprintf(f, "        g12 = %d;\n", gindex + 256);
						if (group->HasMenu()) {
							fprintf(f, "        g11 = %d;\n", group->GetMenu()->authorIndex);
						}
						fprintf(f, "      } </pre>\n");
					} else if (group->GetFirstSubtitleOn()) {
						fprintf(f, "      <pre> {\n");
						fprintf(f, "        if (g0 == 0) {\n");
						fprintf(f, "          g0 = 1;\n");
						fprintf(f, "          subtitle = 64;\n");
						fprintf(f, "        }\n");
						fprintf(f, "      } </pre>\n");
					}
					String vobFileName = String(m_Project->GetOutputDirectory() + S("\\set") + String(gindex + 1) 
						+ S("title") + String(tindex + 1) + S(".mpg")).toFullPath();
					AddFileToDelete(vobFileName);
					String chapters = CreateChapters(gindex, tindex);
					fprintf(f, "      <vob file=\"%s\" chapters=\"%s\" pause=\"2\"/>\n", 
						vobFileName.xmlenc().toUtf8().c_str(), chapters.xmlenc().toUtf8().c_str());
					bool stop = false;
					if (group->HasMenu()) {
						fprintf(f, "      <post> {\n");
						fprintf(f, "        g12 = %d;\n", gindex);
						if (group->HasMenu()) {
							fprintf(f, "        call vmgm menu;\n");
						} else if (tindex < group->GetTitleCount() - 1) {
							fprintf(f, "        g11 = 0;\n");
							fprintf(f, "        g10 = %d;\n", group->GetTitle(tindex + 1)->authorIndex);
							fprintf(f, "        call vmgm menu;\n");
						} else if (gindex < m_Project->GetGroupCount() - 1) {
							if (m_Project->GetGroup(gindex + 1)->HasMenu()) {
								fprintf(f, "        g11 = %d;\n", m_Project->GetGroup(gindex + 1)->GetMenu()->authorIndex);
								fprintf(f, "        call vmgm menu;\n");
							} else {
								fprintf(f, "        g11 = 0;\n");
								fprintf(f, "        g10 = %d;\n", m_Project->GetGroup(gindex + 1)->GetTitle(0)->authorIndex);
								fprintf(f, "        call vmgm menu;\n");
							}
						} else {
							fprintf(f, "    g10 = 1;\n");
							fprintf(f, "    g11 = %d;\n", (m_Project->GetGroup(0)->HasMenu() ? 1 : 0));
						}
						fprintf(f, "      } </post>\n");
					}
					fprintf(f, "    </pgc>\n");
					fprintf(f, "  </titles>\n");
					fprintf(f, "</titleset>\n\n");
				}
			}
		}

		for (int gindex = 0; gindex < m_Project->GetGroupCount(); gindex++) {
			GroupPtr group = m_Project->GetGroup(gindex);
			if (group->HasMenu()) {
				fprintf(f, "<titleset>\n");
				fprintf(f, "  <menus>\n");
				fprintf(f, "    <video aspect=%s/>\n", group->GetMenuAspectRatioStr().enclose().toAnsi().c_str()); 
				fprintf(f, "    <pgc entry=\"root\" pause=\"inf\">\n");
				fprintf(f, "      <pre> {\n");
				SwitchAudioSubtitle(f, gindex);
				fprintf(f, "        button = (g%d %% 32) * 1024;\n", gindex * 2);
				fprintf(f, "      } </pre>\n");
				String vobFileName = String(m_Project->GetOutputDirectory() + S("\\set") + String(gindex + 1) 
					+ S("menu") + String(MENU_ROOT) + S(".mpg")).toFullPath();
				AddFileToDelete(vobFileName);
				fprintf(f, "      <vob file=\"%s\" pause=\"inf\"/>\n", vobFileName.xmlenc().toUtf8().c_str());
				MenuPtr menu = group->GetMenu();
				for (int mindex = 0; mindex < menu->GetMenuItemCount(); mindex++) {
					MenuItemPtr menuItem = menu->GetMenuItem(mindex);
					fprintf(f, "      <button name=\"b%d\"> {\n", mindex);
					fprintf(f, "        jump menu %d;\n", menuItem->authorIndex);
					fprintf(f, "      } </button>\n");
				}
				fprintf(f, "    </pgc>\n");
				for (int mindex = 0; mindex < menu->GetMenuItemCount(); mindex++) {
					MenuItemPtr menuItem = menu->GetMenuItem(mindex);
					fprintf(f, "    <pgc pause=\"0\">\n");
					fprintf(f, "      <pre> {\n");
					fprintf(f, "        g%d = g%d - (g%d %% 32) + %d;\n",
						gindex * 2, gindex * 2, gindex * 2, mindex + 1);
					switch (menuItem->GetSelector1()) {
					case ROOTMENU_PLAYTITLE:
						fprintf(f, "        if ((g10 == %d) and (g12 ge 256)) { resume; }\n",
							group->GetTitle(menuItem->GetSelector2())->authorIndex);
						fprintf(f, "        g10 = %d;\n", group->GetTitle(menuItem->GetSelector2())->authorIndex);
						fprintf(f, "        g11 = 0;\n");
						break;
					case ROOTMENU_GOTOCOMBINED:
						fprintf(f, "        g11 = %d;\n", group->GetCombinedMenu()->authorIndex);
						break;
					case ROOTMENU_GOTOAUDIO:
						fprintf(f, "        g11 = %d;\n", group->GetAudioMenu()->authorIndex);
						break;
					case ROOTMENU_GOTOSUBTITLE:
						fprintf(f, "        g11 = %d;\n", group->GetSubtitleMenu()->authorIndex);
						break;
					case ROOTMENU_GOTOTITLESET:
						if (m_Project->GetGroup(menuItem->GetSelector2())->GetTitleCount()) {
							fprintf(f, "        g10 = %d;\n", m_Project->GetGroup(menuItem->GetSelector2())->GetTitle(0)->authorIndex);
						}
						if (m_Project->GetGroup(menuItem->GetSelector2())->HasMenu()) {
							fprintf(f, "        g11 = %d;\n", m_Project->GetGroup(menuItem->GetSelector2())->GetMenu()->authorIndex);
						} else {
							fprintf(f, "        g11 = 0;\n");
						}
						break;
					case ROOTMENU_RESUME:
						fprintf(f, "        if (g12 ge 256) { resume; }\n");
						if (group->GetTitleCount()) {
							fprintf(f, "        g11 = 0;\n");
						}
						break;
					}
					fprintf(f, "        jump vmgm menu entry title;\n");
					fprintf(f, "      } </pre>\n");
					fprintf(f, "    </pgc>\n");
				}
				if (group->GetMenuType() == MENUTYPE_COMBINED) {
					MenuPtr menu = group->GetCombinedMenu();
					fprintf(f, "    <pgc entry=\"audio,subtitle\" pause=\"inf\">\n");
					fprintf(f, "      <pre> {\n");
					SwitchAudioSubtitle(f, gindex);
					fprintf(f, "        button = ((g%d / 32) %% 32) * 1024;\n", gindex * 2);
					fprintf(f, "      } </pre>\n");
					String vobFileName = String(m_Project->GetOutputDirectory() + S("\\set") + String(gindex + 1) 
						+ S("menu") + String(MENU_COMBINED) + S(".mpg")).toFullPath();
					AddFileToDelete(vobFileName);
					fprintf(f, "      <vob file=\"%s\" pause=\"inf\"/>\n", vobFileName.xmlenc().toUtf8().c_str());
					for (int mindex = 0; mindex < menu->GetMenuItemCount(); mindex++) {
						MenuItemPtr menuItem = menu->GetMenuItem(mindex);
						fprintf(f, "      <button name=\"b%d\"> {\n", mindex);
						fprintf(f, "        jump menu %d;\n", menuItem->authorIndex);
						fprintf(f, "      } </button>\n");
					}
					fprintf(f, "    </pgc>\n");
					for (int mindex = 0; mindex < menu->GetMenuItemCount(); mindex++) {
						MenuItemPtr menuItem = menu->GetMenuItem(mindex);
						fprintf(f, "    <pgc pause=\"0\">\n");
						fprintf(f, "      <pre> {\n");
						fprintf(f, "        g%d = g%d %% 32 + %d + (g%d / 1024) * 1024;\n", 
							gindex * 2, gindex * 2, 32 * (mindex + 1), gindex * 2);
						fprintf(f, "        audio = %d;\n", menuItem->GetSelector1());
						fprintf(f, "        subtitle = %d;\n", (menuItem->GetSelector2() ? menuItem->GetSelector2() + 63 : 0));
						fprintf(f, "        g11 = %d;\n", group->GetMenu()->authorIndex);
						if (menu->GetAutoResume()) {
							fprintf(f, "        if (g12 ge 256) { resume; }\n");
						}
						fprintf(f, "        jump vmgm menu entry title;\n");
						fprintf(f, "      } </pre>\n");
						fprintf(f, "    </pgc>\n");
					}

				} else if (group->GetMenuType() == MENUTYPE_SEPARATE) {
					if (group->HasAudioMenu()) {
						MenuPtr menu = group->GetAudioMenu();
						fprintf(f, "    <pgc entry=\"audio\" pause=\"inf\">\n");
						fprintf(f, "      <pre> {\n");
						SwitchAudioSubtitle(f, gindex);
						fprintf(f, "        button = ((g%d / 32) %% 32) * 1024;\n", gindex * 2);
						fprintf(f, "      } </pre>\n");
						String vobFileName = String(m_Project->GetOutputDirectory() + S("\\set") + String(gindex + 1) 
							+ S("menu") + String(MENU_AUDIO) + S(".mpg")).toFullPath();
						AddFileToDelete(vobFileName);
						fprintf(f, "      <vob file=\"%s\" pause=\"inf\"/>\n", vobFileName.xmlenc().toUtf8().c_str());
						for (int mindex = 0; mindex < menu->GetMenuItemCount(); mindex++) {
							MenuItemPtr menuItem = menu->GetMenuItem(mindex);
							fprintf(f, "      <button name=\"b%d\"> {\n", mindex);
							fprintf(f, "        jump menu %d;\n", menuItem->authorIndex);
							fprintf(f, "      } </button>\n");
						}
						fprintf(f, "    </pgc>\n");
						for (int mindex = 0; mindex < menu->GetMenuItemCount(); mindex++) {
							MenuItemPtr menuItem = menu->GetMenuItem(mindex);
							fprintf(f, "    <pgc pause=\"0\">\n");
							fprintf(f, "      <pre> {\n");
							fprintf(f, "        audio = %d;\n", menuItem->GetSelector1());
							fprintf(f, "        g%d = g%d %% 32 + %d + (g%d / 1024) * 1024;\n", 
								gindex * 2, gindex * 2, 32 * (mindex + 1), gindex * 2);
							fprintf(f, "        g11 = %d;\n", group->GetMenu()->authorIndex);
							if (menu->GetAutoResume()) {
								fprintf(f, "        if (g12 ge 256) { resume; }\n");
							}
							fprintf(f, "        jump vmgm menu entry title;\n");
							fprintf(f, "      } </pre>\n");
							fprintf(f, "    </pgc>\n");
						}
					}
					if (group->HasSubtitleMenu()) {
						MenuPtr menu = group->GetSubtitleMenu();
						fprintf(f, "    <pgc entry=\"subtitle\" pause=\"inf\">\n");
						fprintf(f, "      <pre> {\n");
						SwitchAudioSubtitle(f, gindex);
						fprintf(f, "        button = (g%d / 1024) * 1024;\n", gindex * 2);
						fprintf(f, "      } </pre>\n");
						String vobFileName = String(m_Project->GetOutputDirectory() + S("\\set") + String(gindex + 1) 
							+ S("menu") + String(MENU_SUBTITLE) + S(".mpg")).toFullPath();
						AddFileToDelete(vobFileName);
						fprintf(f, "      <vob file=\"%s\" pause=\"inf\"/>\n", vobFileName.xmlenc().toUtf8().c_str());
						for (int mindex = 0; mindex < menu->GetMenuItemCount(); mindex++) {
							MenuItemPtr menuItem = menu->GetMenuItem(mindex);
							fprintf(f, "      <button name=\"b%d\"> {\n", mindex);
							fprintf(f, "        jump menu %d;\n", menuItem->authorIndex);
							fprintf(f, "      } </button>\n");
						}
						fprintf(f, "    </pgc>\n");
						for (int mindex = 0; mindex < menu->GetMenuItemCount(); mindex++) {
							MenuItemPtr menuItem = menu->GetMenuItem(mindex);
							fprintf(f, "    <pgc pause=\"0\">\n");
							fprintf(f, "      <pre> {\n", mindex);
							fprintf(f, "        subtitle = %d;\n", (menuItem->GetSelector2() ? menuItem->GetSelector2() + 63 : 0));
							fprintf(f, "        g%d = g%d %% 1024 + %d;\n", gindex * 2, gindex * 2, 1024 * (mindex + 1));
							fprintf(f, "        g11 = %d;\n", group->GetMenu()->authorIndex);
							if (menu->GetAutoResume()) {
								fprintf(f, "        if (g12 ge 256) { resume; }\n");
							}
							fprintf(f, "        jump vmgm menu entry title;\n");
							fprintf(f, "      } </pre>\n");
							fprintf(f, "    </pgc>\n");
						}
					}
				}
				fprintf(f, "  </menus>\n");
				fprintf(f, "  <titles>\n");
				fprintf(f, "    <video aspect=%s/>\n", group->GetMenuAspectRatioStr().enclose().toAnsi().c_str()); 
				for (int aindex = 0; aindex < group->GetAudioCount(); aindex++) {
					if (group->GetAudioLanguage(aindex)) {
						fprintf(f, "    <audio lang=\"%s\"/>\n", 
							String(LanguageCode[group->GetAudioLanguage(aindex)]).xmlenc().toUtf8().c_str());
					} else {
						fprintf(f, "    <audio/>\n");
					}
				}
				for (int sindex = 0; sindex < group->GetAudioCount(); sindex++) {
					if (group->GetSubtitleLanguage(sindex)) {
						fprintf(f, "    <subpicture lang=\"%s\"/>\n", 
							String(LanguageCode[group->GetSubtitleLanguage(sindex)]).xmlenc().toUtf8().c_str());
					} else {
						fprintf(f, "    <subpicture/>\n");
					}
				}
				fprintf(f, "    <pgc>\n");
				fprintf(f, "      <pre> { call menu entry root; } </pre>\n", group->GetMenu()->authorIndex);
				fprintf(f, "      <vob file=\"%s\" pause=\"0\"/>\n", tools.blank[m_Project->GetDVDFormat()].xmlenc().toUtf8().c_str());
				fprintf(f, "    </pgc>\n");
				fprintf(f, "  </titles>\n");
				fprintf(f, "</titleset>\n\n");
			}
		}

		fprintf(f, "</dvdauthor>\n");
		fclose(f);
	}

	String logFileName = String(m_Project->GetOutputDirectory() + S("\\") + S("log.txt")).toFullPath();
	String dvdauthorCommand = 
		S("-o ") + String(m_Project->GetOutputDirectory() + S("\\DVD")).enclose()
		+ S(" -x ") + xmlFileName.enclose();
	ok = ok && ExecuteCommand(tools.dvdauthor.c_str(), dvdauthorCommand.c_str(), 
		NULL, NULL, logFileName.c_str(), isCancelled, waitingForUser);

	if (ok) DeleteFilesNow();
	return ok;
}
