#include "menu.h"
#include "group.h"
#include "project.h"
#include "config.h"

#include <algorithm>

Menu::Menu(Group * group, int type) : ManagablePtr()
{
	m_Group           = group;
	m_Type            = type;
	m_BackgroundColor = g_Config->GetMenuColor(MENUCOLOR_BACKGROUND);
	m_NormalColor     = g_Config->GetMenuColor(MENUCOLOR_NORMAL);
	m_SelectColor     = g_Config->GetMenuColor(MENUCOLOR_SELECT);
	m_HiliteColor     = g_Config->GetMenuColor(MENUCOLOR_HILITE);
	m_AutoResume      = true;
	m_SelectedCount   = 0;
	m_FirstSelected   = -1;
}

Menu::~Menu()
{
}

Group * Menu::GetGroup() const
{
	return m_Group;
}

int Menu::GetType() const
{
	return m_Type;
}

const String & Menu::GetBackgroundImage() const
{
	return m_BackgroundImage;
}

bool Menu::SetBackgroundImage(const String & backgroundImage, bool setChanged)
{
	String fullPath = (backgroundImage.length() ? backgroundImage.toFullPath() : backgroundImage);
	if (m_BackgroundImage != fullPath) {
		m_BackgroundImage = fullPath;
		if (setChanged) SetChanged();
	}
	return true;
}

COLORREF Menu::GetBackgroundColor() const
{
	return m_BackgroundColor;
}

bool Menu::SetBackgroundColor(COLORREF backgroundColor, bool setChanged)
{
	if (m_BackgroundColor != backgroundColor) {
		m_BackgroundColor = backgroundColor;
		if (setChanged) SetChanged();
	}
	return true;
}

COLORREF Menu::GetNormalColor() const
{
	return m_NormalColor;
}

bool Menu::SetNormalColor(COLORREF normalColor, bool setChanged)
{
	if (m_NormalColor != normalColor) {
		m_NormalColor = normalColor;
		if (setChanged) SetChanged();
	}
	return true;
}

COLORREF Menu::GetSelectColor() const
{
	return m_SelectColor;
}

bool Menu::SetSelectColor(COLORREF selectColor, bool setChanged)
{
	if (m_SelectColor != selectColor) {
		m_SelectColor = selectColor;
		if (setChanged) SetChanged();
	}
	return true;
}

COLORREF Menu::GetHiliteColor() const
{
	return m_HiliteColor;
}

bool Menu::SetHiliteColor(COLORREF hiliteColor, bool setChanged)
{
	if (m_HiliteColor != hiliteColor) {
		m_HiliteColor = hiliteColor;
		if (setChanged) SetChanged();
	}
	return true;
}

COLORREF Menu::GetColor(int which) const
{
	switch (which) {
	case MENUCOLOR_NORMAL:
		return GetNormalColor();
	case MENUCOLOR_SELECT:
		return GetSelectColor();
	case MENUCOLOR_HILITE:
		return GetHiliteColor();
	default:
		return GetBackgroundColor();
	}
}

bool Menu::GetAutoResume() const
{
	return m_AutoResume;
}

bool Menu::SetAutoResume(bool autoResume, bool setChanged)
{
	if (m_AutoResume != autoResume) {
		m_AutoResume = autoResume;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Menu::CanAddMenuItem() const
{
	return (m_MenuItems.size() < MAX_MENU_ITEMS);
}

bool Menu::AddMenuItem(MenuItemPtr & menuItem, bool setChanged)
{
	if (CanAddMenuItem()) {
		MenuItemVector::iterator pos = std::upper_bound(m_MenuItems.begin(), m_MenuItems.end(), menuItem);
		m_MenuItems.insert(pos, menuItem);
		if (setChanged) {
			menuItem->SetSelected(true);
			menuItem->SetChanged();
		}
		return true;
	}
	return false;
}

bool Menu::IsMenuItemIndexValid(int index) const
{
	return (0 <= index && index < (int) m_MenuItems.size());
}

int Menu::GetMenuItemCount() const
{
	return (int) m_MenuItems.size();
}

MenuItemPtr Menu::GetMenuItem(int index) const
{
	if (IsMenuItemIndexValid(index)) {
		return m_MenuItems[index];
	}
	return MenuItemPtr();
}

bool Menu::SelectMenuItem(int index, bool selected)
{
	if (IsMenuItemIndexValid(index)) {
		m_MenuItems[index]->SetSelected(selected);
		Recalculate();
		return true;
	}
	return false;
}

bool Menu::ClearSelection()
{
	for (MenuItemVector::iterator menuItemIterator = m_MenuItems.begin();
		menuItemIterator != m_MenuItems.end(); menuItemIterator++) {
		(*menuItemIterator)->SetSelected(false);
	}
	Recalculate();
	return true;
}

int Menu::GetSelectedCount() const
{
	return m_SelectedCount;
}

int Menu::GetFirstSelected() const
{
	return m_FirstSelected;
}

MenuItemPtr Menu::GetSelectedMenuItem() const
{
	return GetMenuItem(m_FirstSelected);
}

bool Menu::IsMenuItemSelected(int index) const
{
	if (IsMenuItemIndexValid(index)) {
		return m_MenuItems[index]->IsSelected();
	}
	return false;
}
bool Menu::CanDeleteSelectedMenuItems() const
{
	return (m_SelectedCount > 0);
}

bool Menu::DeleteSelectedMenuItems(bool setChanged)
{
	if (CanDeleteSelectedMenuItems()) {
		for (MenuItemVector::iterator menuItemIterator = m_MenuItems.begin(); menuItemIterator != m_MenuItems.end();) {
			if ((*menuItemIterator)->IsSelected()) {
				menuItemIterator = m_MenuItems.erase(menuItemIterator);
			} else {
				menuItemIterator++;
			}
		}
		if (setChanged) {
			SetChanged();
		} else {
			Recalculate();
		}
		return true;
	}
	return false;
}

void Menu::SetChanged()
{
	m_Group->SetChanged();
}

String Menu::GetIdent()
{
	switch (GetType()) {
	case MENU_ROOT:
		return S("Root menu of group ") + String(GetGroup()->GetIndex() + 1) + S(": ");
	case MENU_COMBINED:
		return S("Audio + subtitle menu of group ") + String(GetGroup()->GetIndex() + 1) + S(": ");
	case MENU_AUDIO:
		return S("Audio menu of group ") + String(GetGroup()->GetIndex() + 1) + S(": ");
	case MENU_SUBTITLE:
		return S("Subtitle menu of group ") + String(GetGroup()->GetIndex() + 1) + S(": ");
	default:
		return S("");
	}
}

String Menu::Recalculate()
{
	String error;

	m_SelectedCount = 0;
	m_FirstSelected = -1;
	for (int mindex = 0; mindex < GetMenuItemCount(); mindex++) {
		MenuItem * menuItem = GetMenuItem(mindex);
		menuItem->SetIndex(mindex);
		error += menuItem->Recalculate();
		if (menuItem->IsSelected()) {
			if (! m_SelectedCount) {
				m_FirstSelected = mindex;
			}
			m_SelectedCount++;
		}
	}

	return error;
}

bool Menu::Load(csv_file_t csv, int & lineno, int nest)
{
	bool ok = true;

	int field = 1;
	ok = ok && SetBackgroundImage(String().fromUtf8(csv_get_str(csv, lineno, ++field, m_BackgroundImage.toUtf8().c_str())), false);
	ok = ok && SetBackgroundColor(csv_get_int(csv, lineno, ++field, m_BackgroundColor), false);
	ok = ok && SetNormalColor(csv_get_int(csv, lineno, ++field, m_NormalColor), false);
	ok = ok && SetSelectColor(csv_get_int(csv, lineno, ++field, m_SelectColor), false);
	ok = ok && SetHiliteColor(csv_get_int(csv, lineno, ++field, m_HiliteColor), false);
	ok = ok && SetAutoResume(csv_get_int(csv, lineno, ++field, m_AutoResume ? 1 : 0) != 0, false);

	lineno++;
	while (ok && lineno < csv->count && csv_get_int(csv, lineno, 0, 0) >= nest) {
		const char * type = csv_get_str(csv, lineno, 1, "");

		if (! strcmp(type, "menuitem")) {
			MenuItemPtr menuItem = new MenuItem(this);
			if (menuItem->Load(csv, lineno, nest + 1)) {
				ok = ok && AddMenuItem(menuItem, false);
			} else {
				ok = false;
			}

		} else {
			lineno++;
		}
	}

	return ok;
}

void Menu::Save(FILE * f, int nest) const
{
	char * type = "";
	switch (m_Type) {
	case MENU_ROOT:
		type = "rootmenu";
		break;
	case MENU_COMBINED:
		type = "combinedmenu";
		break;
	case MENU_AUDIO:
		type = "audiomenu";
		break;
	case MENU_SUBTITLE:
		type = "subtitlemenu";
		break;
	}

	String backgroundImage = String(GetAllocRelativePath(m_BackgroundImage.c_str(), NULL), true);
	fprintf(f, "%d,%s,%s,%d,%d,%d,%d,%d\n", nest, type,
		backgroundImage.enclose().toUtf8().c_str(),
		m_BackgroundColor,
		m_NormalColor,
		m_SelectColor,
		m_HiliteColor,
		(m_AutoResume ? 1 : 0));

	for (MenuItemVector::const_iterator mii = m_MenuItems.begin(); mii != m_MenuItems.end(); mii++) {
		(*mii)->Save(f, nest + 1);
	}
}
