#include "project.h"
#include "tools.h"
#include "config.h"
#include "winutils/winutils.h"

Project::Project() : ManagablePtr()
{
	m_DVDFormat     = g_Config->GetDVDFormat();
	m_DVDSize       = g_Config->GetDVDSize();
	m_CustomSize    = g_Config->GetCustomSize();
	m_CustomBitRate = g_Config->GetCustomBitRate();
	m_OutputType    = g_Config->GetOutputType();

	m_Length          = 0.0;
	m_MaxVideoBitRate = 0;
	m_VideoBitRate    = 0;

	m_HasChanged = false;
}

Project::~Project()
{
}

bool Project::SetDVDFormat(int dvdFormat, bool setChanged)
{
	if (0 <= dvdFormat && dvdFormat < DVDFORMAT_MAX) {
		if (m_DVDFormat != dvdFormat) {
			m_DVDFormat = dvdFormat;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

int Project::GetDVDFormat() const
{
	return m_DVDFormat;
}

bool Project::SetDVDSize(int dvdSize, bool setChanged)
{
	if (0 <= dvdSize && dvdSize < DVDSIZE_MAX) {
		if (m_DVDSize != dvdSize) {
			m_DVDSize = dvdSize;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

int Project::GetDVDSize() const
{
	return m_DVDSize;
}

bool Project::SetCustomSize(int customSize, bool setChanged)
{
	if (0 < customSize) {
		if (m_CustomSize != customSize) {
			m_CustomSize = customSize;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return true;
}

int Project::GetCustomSize() const
{
	return m_CustomSize;
}

bool Project::SetCustomBitRate(int customBitRate, bool setChanged)
{
	if (0 < customBitRate) {
		if (customBitRate > 9000) customBitRate = 9000;
		if (m_CustomBitRate != customBitRate) {
			m_CustomBitRate = customBitRate;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return true;
}

int Project::GetCustomBitRate() const
{
	return m_CustomBitRate;
}

bool Project::SetOutputType(int outputType, bool setChanged)
{
	if (0 <= outputType && outputType < OUTPUTTYPE_MAX) {
		if (m_OutputType != outputType) {
			m_OutputType = outputType;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

int Project::GetOutputType() const
{
	return m_OutputType;
}

bool Project::SetOutputDirectory(const String & outputDirectory, bool setChanged)
{
	String fullPath = outputDirectory.toFullPath();
	if (m_OutputDirectory != fullPath) {
		m_OutputDirectory = fullPath;
		if (setChanged) SetChanged();
	}
	return true;
}

const String & Project::GetOutputDirectory() const
{
	return m_OutputDirectory;
}

bool Project::SetOpenFileDirectory(const String & openFileDirectory, bool setChanged)
{
	String fullPath = openFileDirectory.toFullPath();
	if (m_OpenFileDirectory != fullPath) {
		m_OpenFileDirectory = fullPath;
		if (setChanged) SetChanged();
	}
	return true;
}

const String & Project::GetOpenFileDirectory() const
{
	return m_OpenFileDirectory;
}

int Project::GetGroupCount() const
{
	return (int) m_Groups.size();
}

bool Project::IsGroupIndexValid(int index) const
{
	return (0 <= index && index < (int) m_Groups.size());
}

GroupPtr Project::GetGroup(int index) const
{
	if (IsGroupIndexValid(index)) {
		return m_Groups[index];
	} else {
		return GroupPtr();
	}
}

bool Project::CanAddGroup() const
{
	return (m_Groups.size() < MAX_GROUP_COUNT);
}

bool Project::AddGroup(GroupPtr group, bool setChanged)
{
	if (CanAddGroup()) {
		m_Groups.push_back(group);
		if (setChanged) SetChanged();
		return true;
	}
	return false;
}

bool Project::CanMoveGroupUp(int index) const
{ 
	return (0 < index && index < (int) m_Groups.size());
}

bool Project::MoveGroupUp(int index, bool setChanged)
{
	if (CanMoveGroupUp(index)) {
		GroupPtr group = m_Groups[index];
		m_Groups[index] = m_Groups[index - 1];
		m_Groups[index - 1] = group;
		if (setChanged) SetChanged();
		return true;
	}
	return false;
}

bool Project::CanMoveGroupDown(int index) const
{
	return (0 <= index && index < (int) m_Groups.size() - 1);
}

bool Project::MoveGroupDown(int index, bool setChanged)
{
	if (CanMoveGroupDown(index)) {
		GroupPtr group = m_Groups[index];
		m_Groups[index] = m_Groups[index + 1];
		m_Groups[index + 1] = group;
		if (setChanged) SetChanged();
		return true;
	}
	return false;
}

bool Project::CanDeleteGroup(int index) const
{
	return (0 <= index && index < (int) m_Groups.size());
}

bool Project::DeleteGroup(int index, bool setChanged)
{
	if (CanDeleteGroup(index)) {
		m_Groups.erase(m_Groups.begin() + index);
		if (setChanged) SetChanged();
		return true;
	}
	return false;
}

bool Project::HasChanged() const
{
	return m_HasChanged;
}

void Project::SetChanged()
{
	m_HasChanged = true;
	Recalculate();
}

void Project::ClearChanged()
{
	m_HasChanged = false;
}

void Project::Recalculate()
{
	m_Error.clear();
	m_Length = 0.0;
	double menuLength = 0.0;
	double capacityInMB = (GetDVDSize() == DVDSIZE_BITRATE ? 0.0 : 1.0 * GetDVDCapacity()) - 0.12;
	double subStreamsInKbits = 0.0;
	m_MaxVideoBitRate = 9600;
	bool allHasMenu = true;
	for (int gindex = 0; gindex < GetGroupCount(); gindex++) {
		Group * group = GetGroup(gindex);
		group->SetIndex(gindex);
		m_Error += group->Recalculate();
		m_Length += group->GetLength();
		if (group->HasMenu()) {
			m_Error += group->GetMenu()->Recalculate();
			menuLength += 1.0;
			capacityInMB -= 0.12;
			if (group->GetMenuType() == MENUTYPE_COMBINED) {
				m_Error += group->GetCombinedMenu()->Recalculate();
				menuLength += 1.0;
				capacityInMB -= 0.12;
			} else if (group->GetMenuType() == MENUTYPE_SEPARATE) {
				if (group->HasAudioMenu()) {
					m_Error += group->GetAudioMenu()->Recalculate();
					menuLength += 1.0;
					capacityInMB -= 0.12;
				}
				if (group->HasSubtitleMenu()) {
					m_Error += group->GetSubtitleMenu()->Recalculate();
					menuLength += 1.0;
					capacityInMB -= 0.12;
				}
			}
			capacityInMB -= 0.12;
		} else {
			allHasMenu = false;
		}
		capacityInMB -= 0.12 * group->GetTitleCount();
		subStreamsInKbits += 1.0 * group->GetLength() * group->GetSubStreamRate();
		if (m_MaxVideoBitRate > 10000 - group->GetSubStreamRate()) {
			m_MaxVideoBitRate = 10000 - group->GetSubStreamRate();
		}
	}
	if (GetDVDSize() == DVDSIZE_BITRATE) {
		m_VideoBitRate = m_CustomBitRate;
		m_SizeFromBitRate = (int) (0.5 + (((m_Length + menuLength) * m_VideoBitRate + subStreamsInKbits) / 0.96) * 1000.0 / 8.0 / 1024.0 / 1024.0 - capacityInMB);
	} else if ((m_Length + menuLength) > 0.0) {
		double capacityInKbits = 1024.0 * 1024.0 * 8.0 * capacityInMB / 1000.0;
		capacityInKbits *= 0.96; // muxing overhead
		capacityInKbits -= subStreamsInKbits;
		m_VideoBitRate = (int) (capacityInKbits / (m_Length + menuLength));
	} else {
		m_VideoBitRate = 9000;
	}
	if (m_VideoBitRate > m_MaxVideoBitRate) {
		m_VideoBitRate = m_MaxVideoBitRate;
	}
	if (m_VideoBitRate > 9000) {
		m_VideoBitRate = 9000;
	}

	if (! GetGroupCount()) {
		m_Error += S("The project is empty.\n");
	} else if (GetGroupCount() > 1 && ! allHasMenu) {
		m_Error += S("When there are multiple title groups, all of them must have a menu.\n");
	}
	if (m_VideoBitRate < 2000) {
		m_Error += S("The project is too large to fit on a DVD.\n");
	}
	if (! GetOutputDirectory().length()) {
		m_Error += S("The output directory is not given.\n");
	}
}

const String & Project::GetError() const
{
	return m_Error;
}

String Project::GetVideoFormat() const 
{ 
	return String(m_DVDFormat == DVDFORMAT_PAL ? TEXT("pal") : TEXT("ntsc")); 
}

int Project::GetVideoWidth() const 
{ 
	return (m_DVDFormat == DVDFORMAT_PAL ? 720 : 720); 
};

int Project::GetVideoHeight() const 
{ 
	return (m_DVDFormat == DVDFORMAT_PAL ? 576 : 480); 
};

double Project::GetVideoWHRatio() const
{
	return 1.0 * GetVideoWidth() / GetVideoHeight();
}

double Project::GetAspectCorr(int aspect) const
{
	return GetVideoWHRatio() / ((aspect == ASPECTRATIO_4_3) ? (4.0 / 3.0) : (16.0 / 9.0));
}

double Project::GetFrameRate() const 
{ 
	return (m_DVDFormat == DVDFORMAT_PAL ? 25.0 : (30000.0 / 1001.0)); 
};

int Project::GetDVDCapacity() const
{
	if (m_DVDSize == DVDSIZE_CUSTOM) {
		return (int) m_CustomSize;
	} else {
		return (m_DVDSize == DVDSIZE_DVD_5 ? 4482 : 8148);
	}
};

double Project::GetLength() const
{
	return m_Length;
}

int Project::GetMaxVideoBitRate() const
{
	return m_MaxVideoBitRate;
}

int Project::GetVideoBitRate() const
{
	return m_VideoBitRate;
}

String Project::GetShortDescr() const
{
	String s = S("Sum length: ") + String().fromLength(m_Length) + S("\n");
	if (GetDVDSize() == DVDSIZE_BITRATE) {
		s += S("Estimated size: ") + String(m_SizeFromBitRate) + S("MB");
	} else {
		s += S("Video bitrate: ") + String(m_VideoBitRate) + S(" kbit/s");
	}
	return s;
}

bool Project::Load(csv_file_t csv, const String & directory)
{
	LPTSTR origDir = ChangeToDirectory(directory.c_str());

	bool ok = ! csv_get_int(csv, 0, 0, 1) && ! strcmp(csv_get_str(csv, 0, 1, ""), "DVDforger")
		&& (csv_get_int(csv, 1, 0, 0) == 1) && ! strcmp(csv_get_str(csv, 1, 1, ""), "project");
	if (ok) {
		
		COLORREF colors[16];
		bool gotColors = false;
		
		m_Groups.clear();
		
		int lineno = 1;
		int field = 1;
		ok = ok && SetDVDFormat(csv_get_int(csv, lineno, ++field, m_DVDFormat), false);
		ok = ok && SetDVDSize(csv_get_int(csv, lineno, ++field, m_DVDSize), false);
		ok = ok && SetCustomSize(csv_get_int(csv, lineno, ++field, m_CustomSize), false);
		int outputType = csv_get_int(csv, lineno, ++field, m_OutputType - OUTPUTTYPE_DVDFOLDER) + OUTPUTTYPE_DVDFOLDER;
		ok = ok && SetOutputDirectory(String().fromUtf8(csv_get_str(csv, lineno, ++field, m_OutputDirectory.toUtf8().c_str())), false);
		ok = ok && SetOpenFileDirectory(String().fromUtf8(csv_get_str(csv, lineno, ++field, m_OpenFileDirectory.toUtf8().c_str())), false);
		ok = ok && SetOutputType(csv_get_int(csv, lineno, ++field, outputType), false);
		ok = ok && SetCustomBitRate(csv_get_int(csv, lineno, ++field, m_CustomBitRate), false);
		ok = ok && SetDVDSize(csv_get_int(csv, lineno, ++field, m_DVDSize), false);

		lineno++;
		int nest = 2;
		while (ok && lineno < csv->count && csv_get_int(csv, lineno, 0, 0) == nest) {
			const char * type = csv_get_str(csv, lineno, 1, "");
			
			if (! strcmp(type, "group")) {
				GroupPtr group = new Group(this);
				if (group->Load(csv, lineno, nest + 1)) {
					ok = AddGroup(group, false);
				} else {
					ok = false;
				}
			
			} else if (! strcmp(type, "colors")) {
				gotColors = true;
				for (int i = 0; i < 16; i++) {
					colors[i] = csv_get_int(csv, lineno, i + 2, 0);
				}
				lineno++;
			
			} else {
				lineno++;
			}
		}

		if (lineno < csv->count) ok = false;

		if (ok && gotColors) {
			for (int i = 0; i < 16; i++) {
				CustomColors[i] = colors[i];
			}
		}
	}

	ChangeDirectoryBack(origDir);

	return ok;
}

void Project::Save(FILE * f, const String & directory) const
{
	LPTSTR origDir = ChangeToDirectory(directory.c_str());

	String outputDirectory   = String(GetAllocRelativePath(m_OutputDirectory.c_str(),   NULL), true);
	String openFileDirectory = String(GetAllocRelativePath(m_OpenFileDirectory.c_str(), NULL), true);
	fprintf(f, "0,DVDforger\n");
	fprintf(f, "1,project,%d,%d,%d,%d,%s,%s,%d,%d,%d\n",
		m_DVDFormat, m_DVDSize % DVDSIZE_BITRATE, m_CustomSize, 
		(m_OutputType < OUTPUTTYPE_DVDFOLDER ? 0 : m_OutputType - OUTPUTTYPE_DVDFOLDER),
		outputDirectory.enclose(TEXT('"')).toUtf8().c_str(),
		openFileDirectory.enclose(TEXT('"')).toUtf8().c_str(),
		m_OutputType, m_CustomBitRate, m_DVDSize);

	for (GroupVector::const_iterator groupIterator = m_Groups.begin(); 
		groupIterator != m_Groups.end(); groupIterator++) {
		(*groupIterator)->Save(f, 2);
	}

	fprintf(f, "2,colors");
	for (int i = 0; i < 16; i++) {
		fprintf(f, ",%d", CustomColors[i]);
	}
	fprintf(f, "\n");

	ChangeDirectoryBack(origDir);
}

