#include "subtitle.h"
#include "langcode.h"
#include "video.h"
#include "title.h"
#include "group.h"
#include "project.h"
#include "subtitlereader.h"
#include "config.h"

#include "winutils/winutils.h"

Subtitle::Subtitle(Video * video, Subtitle * subtitle) : ManagablePtr()
{
	m_Video = video;
	if (subtitle) {
		m_FileName      = subtitle->m_FileName;
		m_Charset       = -1;
		m_FontFace      = subtitle->m_FontFace;
		m_FontSize      = subtitle->m_FontSize;
		m_Bold          = subtitle->m_Bold;
		m_Italic        = subtitle->m_Italic;
		m_Box           = subtitle->m_Box;
		m_Color         = subtitle->m_Color;
		m_Advance       = subtitle->m_Advance;
		m_Offset        = subtitle->m_Offset;
		m_IsChecked     = subtitle->m_IsChecked;
		m_IsValid       = subtitle->m_IsValid;
		m_FileSelected  = subtitle->m_FileSelected;
		m_TrackSelected = subtitle->m_TrackSelected;
	} else {
		m_Charset       = -1;
		m_FontFace      = g_Config->GetFontFace();
		m_FontSize      = g_Config->GetFontSize();
		m_Bold          = g_Config->GetBold();
		m_Italic        = g_Config->GetItalic();
		m_Box           = g_Config->GetBox();
		m_Color         = g_Config->GetColor();
		m_Advance       = false;
		m_Offset        = 0;
		m_IsChecked     = false;
		m_IsValid       = false;
		m_FileSelected  = false;
		m_TrackSelected = false;
	}
	m_Index = -1;
}

Subtitle::~Subtitle()
{
}

Video * Subtitle::GetVideo() const
{
	return m_Video;
}

bool Subtitle::IsFileSelected() const
{
	return m_FileSelected;
}


void Subtitle::SetFileSelected(bool selected)
{
	m_FileSelected = selected;
}

bool Subtitle::IsTrackSelected() const
{
	return m_TrackSelected;
}

void Subtitle::SetTrackSelected(bool selected)
{
	m_TrackSelected = selected;
}

const String & Subtitle::GetFileName() const
{
	return m_FileName;
}

bool Subtitle::SetFileName(const String & fileName, bool setChanged)
{
	String fullPath = fileName.toFullPath();
	if (m_FileName != fullPath) {
		m_FileName = fullPath;
		if (setChanged) SetChanged();
		m_IsChecked = false;
		m_IsValid   = false;
	}
	return true;
}

int Subtitle::GetCharset() const
{
	return m_Charset;
}

bool Subtitle::SetCharset(int charset, bool setChanged) 
{
	if (0 <= charset && charset < CHARSET_COUNT) {
		if (m_Charset != charset) {
			m_Charset = charset;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

const String & Subtitle::GetFontFace() const
{
	return m_FontFace;
}

bool Subtitle::SetFontFace(const String & fontFace, bool setChanged)
{
	if (m_FontFace != fontFace) {
		m_FontFace = fontFace;
		if (setChanged) SetChanged();
	}
	return true;
}

int Subtitle::GetFontSize() const
{
	return m_FontSize;
}

bool Subtitle::SetFontSize(int fontSize, bool setChanged)
{
	if (0 < fontSize) {
		if (m_FontSize != fontSize) {
			m_FontSize = fontSize;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

bool Subtitle::GetBold() const
{
	return m_Bold;
}

bool Subtitle::SetBold(bool bold, bool setChanged)
{
	if (m_Bold != bold) {
		m_Bold = bold;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Subtitle::GetItalic() const
{
	return m_Italic;
}

bool Subtitle::SetItalic(bool italic, bool setChanged)
{
	if (m_Italic != italic) {
		m_Italic = italic;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Subtitle::GetBox() const
{
	return m_Box;
}

bool Subtitle::SetBox(bool box, bool setChanged)
{
	if (m_Box != box) {
		m_Box = box;
		if (setChanged) SetChanged();
	}
	return true;
}

COLORREF Subtitle::GetColor() const
{
	return m_Color;
}

bool Subtitle::SetColor(COLORREF color, bool setChanged)
{
	if (m_Color != color) {
		m_Color = color;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Subtitle::GetAdvance() const
{
	return m_Advance;
}

bool Subtitle::SetAdvance(bool advance, bool setChanged)
{
	if (m_Advance != advance) {
		m_Advance = advance;
		if (setChanged) SetChanged();
	}
	return true;
}

int Subtitle::GetOffset() const
{
	return m_Offset;
}

bool Subtitle::SetOffset(int offset, bool setChanged)
{
	if (0 <= offset) {
		if (m_Offset != offset) {
			m_Offset = offset;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

int Subtitle::GetBitRateInKbps() const
{
	return 16;	
}

int Subtitle::GetIndex() const
{
	return m_Index;
}

void Subtitle::SetIndex(int index)
{
	m_Index = index;
}

String Subtitle::GetIdent(bool forTrack)
{
	if (m_Index < 0) {
		return S("");
	} else if (forTrack) {
		return S("Subtitle track ") + String(GetIndex() + 1) +
			S(" of title ") + String(GetVideo()->GetTitle()->GetIndex() + 1) +
			S(" in group ") + String(GetVideo()->GetTitle()->GetGroup()->GetIndex() + 1) +
			S(": ");
	} else {
		return S("Subtitle file ") + String(GetIndex() + 1) +
			S(" of subtitle track ") + String(GetVideo()->GetIndex() + 1) +
			S(" of title ") + String(GetVideo()->GetTitle()->GetIndex() + 1) +
			S(" in group ") + String(GetVideo()->GetTitle()->GetGroup()->GetIndex() + 1) +
			S(": ");
	}
}

String Subtitle::Recalculate()
{
	String error;
	if (! IsChecked()) Analyze();
	if (! GetFontInfo(GetFontFace().c_str(), NULL, NULL)) {
		error += (GetIdent(true) + S("There's no such font as ") + GetFontFace().enclose() + S("\n"));
	}
	if (! IsValid() && GetFileName().length()) {
		error += (GetIdent(false) + GetFileName().enclose() + S(" doesn't seem to be a subtitle file.\n"));
	}
	return error;
}

void Subtitle::SetChanged()
{
	m_Video->SetChanged();
}

bool Subtitle::Analyze()
{
	m_IsChecked = true;
	if (m_FileName.length()) {
		if (m_Charset < 0) m_Charset = (IsTextFileUTF8(m_FileName.c_str()) ? 1 : 0);
		SubtitleEntryVector entries;
		SubtitleReader subtitleReader(entries, GetVideo()->GetLength(), GetVideo()->GetGroup()->GetProject()->GetFrameRate());
		m_IsValid = subtitleReader.ReadSubtitles(m_FileName, (m_Charset != 0), GetVideo()->GetOrigFPS(), 0.0, 1.0);
	} else {
		m_IsValid = false;
	}
	Recalculate();
	return m_IsValid;
}

bool Subtitle::IsChecked() const
{
	return m_IsChecked;
}

bool Subtitle::IsValid() const
{
	return m_IsValid;
}

void Subtitle::GetDescription(StringVectorVector & strings, bool forTitle) const
{
	StringVector line;
	if (forTitle) {
		if (GetVideo()->GetGroup()->GetView() == VIEW_TITLETRACKPART) {
			line.push_back(S("Subtitle track ") + String(GetIndex() + 1) + S(": "));
		} else {
			line.push_back(S("Title ") + String(GetVideo()->GetTitle()->GetIndex() + 1) + S(": "));
		}
		String s = GetFontFace() + S(" ") + String(GetFontSize()) + S(" pt");
		if (GetBold()) s += S(", bold");
		if (GetItalic()) s += S(", italic");
		if (GetBox()) s += S(", box");
		line.push_back(s);
		strings.push_back(line);
	} else {
		String s = S("Subtitle file");
		if (GetVideo()->GetTitle()->GetVideoCount() > 1) {
			s += S(" ") + String(GetVideo()->GetIndex() + 1);
		}
		s += S(": ");
		line.push_back(s);
		s = GetFileName().toFileName();
		line.push_back(s);
		strings.push_back(line);
		line.clear();
		line.push_back(String());
		if (s.length()) {
			s = S("Charset: ") + String(CharsetName[GetCharset()]);
			if (GetOffset()) {
				s += ((GetAdvance() ? S(", advance ") : S(", delay ")) + String(GetOffset()) + S(" ms"));
			}
		} else {
			s = S(" ");
		}
		line.push_back(s);
		strings.push_back(line);
	}
}

bool Subtitle::Load(csv_file_t csv, int & lineno, int nest)
{
	bool ok = true;
	
	int field = 1;
	ok = ok && SetFileName(String().fromUtf8(csv_get_str(csv, lineno, ++field, m_FileName.toUtf8().c_str())), false);
    ok = ok && SetCharset(csv_get_int(csv, lineno, ++field, m_Charset), false);
	ok = ok && SetFontFace(String().fromUtf8(csv_get_str(csv, lineno, ++field, m_FontFace.toUtf8().c_str())), false);
	ok = ok && SetFontSize(csv_get_int(csv, lineno, ++field, m_FontSize), false);
	ok = ok && SetBold(csv_get_int(csv, lineno, ++field, m_Bold ? 1 : 0) != 0, false);
	ok = ok && SetItalic(csv_get_int(csv, lineno, ++field, m_Italic ? 1 : 0) != 0, false);
	ok = ok && SetBox(csv_get_int(csv, lineno, ++field, m_Box ? 1 : 0) != 0, false);
	ok = ok && SetColor(csv_get_int(csv, lineno, ++field, m_Color), false);
	ok = ok && SetAdvance(csv_get_int(csv, lineno, ++field, m_Advance ? 1 : 0) != 0, false);
	ok = ok && SetOffset(csv_get_int(csv, lineno, ++field, m_Offset), false);
	SetFileSelected(csv_get_int(csv, lineno, ++field, m_FileSelected ? 1 : 0) != 0);
	SetTrackSelected(csv_get_int(csv, lineno, ++field, m_TrackSelected ? 1 : 0) != 0);

	lineno++;
	while (ok && lineno < csv->count && csv_get_int(csv, lineno, 0, 0) >= nest) {
		lineno++;
	}

	return ok;
}

void Subtitle::Save(FILE * f, int nest) const
{
	String fileName = String(GetAllocRelativePath(m_FileName.c_str(), NULL), true);
	fprintf(f, "%d,subtitle,%s,%d,%s,%d,%d,%d,%d,%d,%d,%d,%d,%d\n", nest,
		fileName.enclose().toUtf8().c_str(),
		m_Charset,
		m_FontFace.enclose().toUtf8().c_str(),
		m_FontSize,
		(m_Bold   ? 1 : 0),
		(m_Italic ? 1 : 0),
		(m_Box    ? 1 : 0),
		m_Color,
		(m_Advance ? 1 : 0),
		m_Offset,
		(m_FileSelected ? 1 : 0),
		(m_TrackSelected ? 1 : 0));
}
