#include "subtitledrawer.h"
#include "group.h"
#include "title.h"
#include "video.h"
#include "subtitle.h"
#include "subtitlereader.h"
#include "subtitleentry.h"
#include "tools.h"

#include "EasyBMP/EasyBMP.h"
#include "EasyBMP/EasyBMP_win32.h"
#include "winutils/winutils.h"
#include "safefgets/safefgets.h"

SubtitleDrawer::SubtitleDrawer(ProjectPtr project, int groupIndex, int titleIndex, int subtitleIndex)
{
	m_Project       = project;
	m_GroupIndex = groupIndex;
	m_TitleIndex    = titleIndex;
	m_SubtitleIndex = subtitleIndex;
}

SubtitleDrawer::~SubtitleDrawer()
{
}

String SubtitleDrawer::GetDescription() const
{
	return String(S("Creating subtitle ") + String(m_SubtitleIndex + 1) 
		+ S(" of title ") + String(m_TitleIndex + 1) + S(" in group ") + String(m_GroupIndex + 1));
}

bool SubtitleDrawer::OnRun()
{
	bool ok = true;

	int maxWidth = m_Project->GetVideoWidth() * 8 / 10;
	int bottom   = m_Project->GetVideoHeight() * 9 / 10;

	TitlePtr title = m_Project->GetGroup(m_GroupIndex)->GetTitle(m_TitleIndex);

	SubtitleEntryVector entries;
	SubtitleReader reader(entries, title->GetLength(), m_Project->GetFrameRate());
	double   offset = 0.0;
	bool     first  = true;
	String   fontFace;
	int      fontSize = 0;
	bool     bold     = false;
	bool     italic   = false;
	bool     box      = false;
	COLORREF color    = RGB(255, 255, 255);
	for (int index = 0; index < title->GetVideoCount(); index++) {
		VideoPtr video = title->GetVideo(index);
		SubtitlePtr subtitle = video->GetSubtitle(m_SubtitleIndex);
		if (subtitle->IsValid()) {
			if (first) {
				fontFace = subtitle->GetFontFace();
				fontSize = subtitle->GetFontSize();
				bold     = subtitle->GetBold();
				italic   = subtitle->GetItalic();
				box      = subtitle->GetBox();
				color    = subtitle->GetColor();
				first = false;
			}
			double thisOffset = offset;
			if (subtitle->GetAdvance()) {
				thisOffset -= (subtitle->GetOffset() / 1000.0);
			} else {
				thisOffset += (subtitle->GetOffset() / 1000.0);
			}
			reader.ReadSubtitles(subtitle->GetFileName(), (subtitle->GetCharset() != 0), 
				video->GetOrigFPS(), thisOffset, video->GetAcceleration());
		}
		offset += video->GetLength();
	}
	SubtitleEntryPtr finalEntry = SubtitleEntryPtr(new SubtitleEntry());
	finalEntry->AppendLine(S(" "));
	finalEntry->SetStartTime(title->GetLength() - 1.0 / m_Project->GetFrameRate());
	finalEntry->SetEndTime(title->GetLength());
	entries.push_back(finalEntry);

	HDC dc = CreateCompatibleDC(NULL);
	LOGFONT lf;
	if (! GetFontInfo(fontFace.c_str(), & lf, NULL)) {
		GetFontInfo(TEXT("Arial"), & lf, NULL);
	}
	double aspectRatio  = title->GetVideo(0)->GetFinalAspectRatio();
	double videoWHRatio = m_Project->GetVideoWHRatio();
	double fontWHRatio  = 1.0 * lf.lfWidth / lf.lfHeight;
	lf.lfHeight    = fontSize;
	lf.lfWidth     = (int) (0.5 + fontSize * fontWHRatio * videoWHRatio / aspectRatio);
	lf.lfWeight    = (bold ? FW_BOLD : FW_NORMAL);
	lf.lfItalic    = italic;
	HFONT font = CreateFontIndirect(& lf);
	SelectObject(dc, font);
	HBRUSH brush = CreateSolidBrush(RGB(1, 1, 1));
	SetTextColor(dc, color);
	SetBkMode(dc, TRANSPARENT);

	String baseFileName = String(m_Project->GetOutputDirectory() + S("\\set") + String(m_GroupIndex + 1) 
		+ S("title") + String(m_TitleIndex + 1)).toFullPath();
	String mpgFileName = baseFileName + S(".mpg");
	baseFileName += (S("subtitle") + String(m_SubtitleIndex + 1));
	String xmlFileName = baseFileName + S(".xml");
	String tmpFileName = baseFileName + S(".mpg");
	String imageFileName = baseFileName + S("image");
	String logFileName = String(m_Project->GetOutputDirectory() + S("\\") + S("log.txt")).toFullPath();

	AddFileToDelete(xmlFileName);

	bool doItAgain = true;
	int adjustEntry = -1;
	while (ok && doItAgain) {
		std::vector<int> xoffsets;
		std::vector<int> yoffsets;
		for (int index = 0; ok && index < (int) entries.size(); index++) {
			if (adjustEntry < 0 || adjustEntry == index) {
				RECT rect;
				if (adjustEntry < 0) {
					bool go = true;
					while (go) {
						Measure(entries[index], dc, rect);
						go = (rect.right > maxWidth);
						if (go) entries[index]->BreakLongestLine();
					}
				} else {
					entries[index]->BreakLongestLine();
					Measure(entries[index], dc, rect);
				}
				int xoffset = (m_Project->GetVideoWidth() - rect.right) / 2;
				int yoffset = (bottom - rect.bottom);
				if (yoffset < 0) yoffset = 0;
				if (adjustEntry < 0) {
					xoffsets.push_back(xoffset);
					yoffsets.push_back(yoffset);
				} else {
					xoffsets[index] = xoffset;
					yoffsets[index] = yoffset;
				}
				HBITMAP bitmap = CreateBitmap(rect.right, rect.bottom, 1, 32, NULL);
				SelectObject(dc, bitmap);
				FillRect(dc, & rect, brush);
				Draw(entries[index], dc, rect, box);
				BMP image;
				HBITMAPtoBMP(dc, bitmap, image);
				DeleteObject(bitmap);
				image.SetBitDepth(4);
				RGBApixel colorTable[16];
				for (int c = 0; c < 16; c++) {
					colorTable[c].Red   = (c < 3 ? GetRValue(color) * c / 2 : 1);
					colorTable[c].Green = (c < 3 ? GetGValue(color) * c / 2 : 1);
					colorTable[c].Blue  = (c < 3 ? GetBValue(color) * c / 2 : 1);
					colorTable[c].Alpha = (c < 3 ? 0 : 255);
					image.SetColor(c, colorTable[c]);
				}
				String bmpFileName = imageFileName + String(index) + S(".bmp");
				image.WriteToFile(bmpFileName.toAnsi().c_str());
				String pngFileName = imageFileName + String(index) + S(".png");
				if (adjustEntry >= 0) DeleteFile(pngFileName.c_str());
				String bmp2pngCommand = S("-X -Q -E -P #010101 ") + bmpFileName.enclose();
				ok = ok && ExecuteCommand(tools.bmp2png.c_str(), bmp2pngCommand.c_str(), 
					NULL, NULL, logFileName.c_str(), isCancelled, waitingForUser);
				if (adjustEntry < 0) AddFileToDelete(pngFileName);
			}
		}
		if (ok) {
			FILE * f = fopen(xmlFileName.toAnsi().c_str(), "wt");
			if (f) {
				fprintf(f, "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n");
				fprintf(f, "<subpictures>\n<stream>\n");
				for (int index = 0; ok && index < (int) entries.size(); index++) {
					fprintf(f, "  <spu start=\"%s\" end=\"%s\" image=\"%s\" "
						"xoffset=\"%d\" yoffset=\"%d\" transparent=\"010101\" force=\"no\"/>\n",
						entries[index]->GetFormattedStartTime().xmlenc().toUtf8().c_str(),
						entries[index]->GetFormattedEndTime().xmlenc().toUtf8().c_str(),
						String(imageFileName + String(index) + S(".png")).xmlenc().toUtf8().c_str(),
						xoffsets[index], yoffsets[index]);
				}
				fprintf(f, "</stream></subpictures>\n");
				fclose(f);
			}
			String spumuxCommand = S("-s ") + String(m_SubtitleIndex) + S(" ") + xmlFileName.enclose();
			doItAgain = ! ExecuteCommand(tools.spumux.c_str(), spumuxCommand.c_str(),
				mpgFileName.c_str(), tmpFileName.c_str(), logFileName.c_str(), isCancelled, waitingForUser);
			if (doItAgain) {
				ok = false;
				if (! * isCancelled) {
					bool headFound = false;
					adjustEntry = -1;
					f = fopen(logFileName.toAnsi().c_str(), "rt");
					if (f) {
						char * line = NULL;
						while ((line = safefgets(line, f)) != NULL) {
							int x;
							char * test = line;
							for (x = 0; line[x]; x++) {
								if (line[x] == 0x0D) test = line + x + 1;
							}
							if (sscanf(test, "DVDAuthor::spumux, version %d", &x) == 1) {
								ok = false;
								headFound = true;
								adjustEntry = -2;
							} else if (headFound) {
								if (sscanf(test, "INFO: PNG had %d colors", &x) == 1) {
									adjustEntry++;
								} else if (sscanf(test, "ERR: Encoded row takes more than %d bits.", &x) == 1) {
									ok = (adjustEntry >= 0);
								}
							}
						}
						fclose(f);
					}
				}
			}
		}
	}
	DeleteObject(font);
	DeleteDC(dc);
	ok = ok && MoveFileEx(tmpFileName.c_str(), mpgFileName.c_str(), MOVEFILE_REPLACE_EXISTING | MOVEFILE_COPY_ALLOWED);
	
	if (ok) DeleteFilesNow();
	return ok;
}

void SubtitleDrawer::Measure(SubtitleEntryPtr entry, HDC dc, RECT & rect)
{
	int width  = 0;
	int height = 4;
	for (int index = 0; index < entry->GetLineCount(); index++) {
		SIZE size;
		GetTextExtentPoint32(dc, entry->GetLine(index).c_str(), (int) entry->GetLine(index).length(), & size);
		height += size.cy;
		if (width < size.cx) width = size.cx;
	}
	width += 4;
	rect.left   = 0;
	rect.top    = 0;
	rect.right  = width;
	rect.bottom = height;
}

void SubtitleDrawer::Draw(SubtitleEntryPtr entry, HDC dc, const RECT & rect, bool box)
{
	int middle = (rect.right - rect.left) / 2;
	int y = rect.top + 2;
	for (int index = 0; index < entry->GetLineCount(); index++) {
		SIZE size;
		GetTextExtentPoint32(dc, entry->GetLine(index).c_str(), (int) entry->GetLine(index).length(), & size);
		int x = middle - size.cx / 2;
		if (box) {
			RECT boxRect;
			boxRect.left   = x - 2;
			boxRect.top    = y - 2;
			boxRect.right  = boxRect.left + size.cx + 4;
			boxRect.bottom = boxRect.top + size.cy + 4;
			FillRect(dc, & boxRect, (HBRUSH) GetStockObject(BLACK_BRUSH));
		} else {
			COLORREF origColor = SetTextColor(dc, RGB(0, 0, 0));
			for (int dx = -2; dx <= 2; dx++) {
				for (int dy = -2; dy <= 2; dy++) {
					TextOut(dc, x + dx, y + dy, entry->GetLine(index).c_str(), (int) entry->GetLine(index).length());
				}
			}
			SetTextColor(dc, origColor);
		}
		TextOut(dc, x, y, entry->GetLine(index).c_str(), (int) entry->GetLine(index).length());
		y += size.cy;
	}
}
