#include "subtitlereader.h"
#include "safefgets/safefgets.h"
#include "tools.h"
#include <algorithm>

#define STATE_LOOKINGFORSOMETHING 0

#define STATE_SRT_INSIDEENTRY 10

#define STATE_SSA_FORMATEXPECTED   20
#define STATE_SSA_DIALOGUEEXPECTED 21

SubtitleReader::SubtitleReader(SubtitleEntryVector & entries, double length, double fps) : m_Entries(entries)
{
	m_Length = 1.0 * ((int)(length * fps) - 1) / fps;
	m_TargetFPS = fps;
}

SubtitleReader::~SubtitleReader()
{
}

bool SubtitleReader::AdjustStartEnd(double & start, double & end)
{
	start = 1.0 * ((int)((m_Offset + m_Acceleration * start) * m_TargetFPS)) / m_TargetFPS;
	end   = 1.0 * ((int)((m_Offset + m_Acceleration * end) * m_TargetFPS)) / m_TargetFPS;
	if (end < start) {
		double t = start;
		start = end;
		end = t;
	}
	if (0.0 < end && start < m_Length) {
		if (start < 0.0) start = 0.0;
		if (end > m_Length) end = m_Length;
		return true;
	} else {
		return false;
	}
}

void SubtitleReader::LookForSomething(char * line)
{
	if (m_State == STATE_LOOKINGFORSOMETHING) { 
		int shour, smin, ssec, smilli, ehour, emin, esec, emilli;
		if ((sscanf(line, "%d:%d:%d,%d --> %d:%d:%d,%d", 
			        & shour, & smin, & ssec, & smilli, 
			        & ehour, & emin, & esec, & emilli) == 8) ||
			(sscanf(line, "%d:%d:%d.%d,%d:%d:%d.%d", 
			        & shour, & smin, & ssec, & smilli, 
			        & ehour, & emin, & esec, & emilli) == 8)) {
			m_FoundSomething = true;
			double start = ((1.0 * shour) * 60.0 + smin) * 60.0 + ssec + (1.0 * smilli / 1000.0);
			double end   = ((1.0 * ehour) * 60.0 + emin) * 60.0 + esec + (1.0 * emilli / 1000.0);
			if (AdjustStartEnd(start, end)) {
				m_State = STATE_SRT_INSIDEENTRY;
				SubtitleEntryPtr entry = SubtitleEntryPtr(new SubtitleEntry());
				entry->SetStartTime(start);
				entry->SetEndTime(end);
				m_Entries.push_back(entry);
				m_LastEntry = entry;
			}
		}
	}

	if (m_State == STATE_LOOKINGFORSOMETHING) { 
		int startFrame, endFrame;
		if (sscanf(line, "{%d}{%d}", & startFrame, & endFrame) == 2) {
			FetchMicroDVDLine(line, startFrame, endFrame);
		}
	}
	
	if (m_State == STATE_LOOKINGFORSOMETHING) { 
		String s = (m_Utf8 ? String().fromUtf8(line).trim() : String().fromAnsi(line).trim());
		if (s == S("[Events]")) {
			m_State = STATE_SSA_FORMATEXPECTED;
		}
	}
}

void SubtitleReader::FetchSrtLine(char * line)
{
	String s = (m_Utf8 ? String().fromUtf8(line).trim() : String().fromAnsi(line).trim());
	if (s.length()) {
		String entryLine;
		int inside = 0;
		String::iterator si = s.begin();
		while (si != s.end()) {
			if ((*si) == TEXT('<') || (*si) == TEXT('{')) {
				inside++;
			} else if ((*si) == TEXT('>') || (*si) == TEXT('}')) {
				inside--;
			} else {
				entryLine += (*si);
			}
			si++;
		}
		m_LastEntry->AppendLine(entryLine);
	} else {
		m_State = STATE_LOOKINGFORSOMETHING;
	}
}

void SubtitleReader::FetchMicroDVDLine(char * line, int startFrame, int endFrame)
{
	String s = (m_Utf8 ? String().fromUtf8(line).trim() : String().fromAnsi(line).trim());
	int pos = 0;
	int bracount = 0;
	while ((pos < (int) s.length()) && (bracount < 4)) {
		if ((s[pos] == TEXT('{')) || (s[pos] == TEXT('}'))) bracount++;
		pos++;
	}
	if ((pos < (int) s.length()) && (bracount == 4)) {
		m_FoundSomething = true;
		double start = 1.0 * startFrame / m_SourceFPS;
		double end   = 1.0 * endFrame   / m_SourceFPS;
		if (AdjustStartEnd(start, end)) {
			SubtitleEntryPtr entry = SubtitleEntryPtr(new SubtitleEntry());
			entry->SetStartTime(start);
			entry->SetEndTime(end);
			m_Entries.push_back(entry);
			String entryLine;
			while (pos < (int) s.length()) {
				if (s[pos] == TEXT('|')) {
					entry->AppendLine(entryLine);
					entryLine.clear();
				} else {
					entryLine += s[pos];
				}
				pos++;
			}
			if (entryLine.length()) {
				entry->AppendLine(entryLine);
			}
		}
	}
}

void SubtitleReader::ProcessSsaFormat(char * line)
{
	String s = (m_Utf8 ? String().fromUtf8(line).trim() : String().fromAnsi(line).trim());
	if (s.substr(0, 7) == S("Format:")) {
		m_SsaStart = 0;
		m_SsaEnd   = 0;
		m_SsaText  = 0;
		int commas = 0;
		int pos = 7;
		while ((pos < (int) s.length()) && iswspace(s[pos])) pos++;
		while (pos < (int) s.length()) {
			// to-do: extract times
			if ((s.substr(pos, 5) == S("Start")) && ((pos + 6 >= (int) s.length()) || iswspace(s[pos + 6]) || (s[pos + 6] == TEXT(',')))) {
				m_SsaStart = commas;
			} else if ((s.substr(pos, 3) == S("End")) && ((pos + 4 >= (int) s.length()) || iswspace(s[pos + 4]) || (s[pos + 4] == TEXT(',')))) {
				m_SsaEnd = commas;
			} else if ((s.substr(pos, 4) == S("Text")) && ((pos + 5 >= (int) s.length()) || iswspace(s[pos + 5]) || (s[pos + 5] == TEXT(',')))) {
				m_SsaText = commas;
				m_State = STATE_SSA_DIALOGUEEXPECTED;
			}
			while ((pos < (int) s.length()) && (s[pos] != TEXT(','))) pos++;
			if (pos < (int) s.length()) {
				commas++;
				pos++;
				while ((pos < (int) s.length()) && iswspace(s[pos])) pos++;
			}
		}
	}
}

void SubtitleReader::ProcessSsaDialogue(char * line)
{
	String s = (m_Utf8 ? String().fromUtf8(line).trim() : String().fromAnsi(line).trim());
	if (s.substr(0, 9) == S("Dialogue:")) {
		double start = 0;
		double end   = 0;
		int commas = 0;
		int pos = 9;
		while ((pos < (int) s.length()) && (commas < m_SsaText)) {
			if (s[pos] == TEXT(',')) {
				commas++;
				pos++;
				while ((pos < (int) s.length()) && iswspace(s[pos])) pos++;
				if (commas == m_SsaStart || commas == m_SsaEnd) {
					double t = 0.0;
					String::const_iterator iter = s.begin() + pos;
					ParseLength(iter, s.end(), t);
					if (commas == m_SsaStart) {
						start = t;
					}
					if (commas == m_SsaEnd) {
						end = t;
					}
				}
			} else {
				pos++;
			}
		}
		if (commas == m_SsaText) {
			m_FoundSomething = true;
			if (AdjustStartEnd(start, end)) {
				SubtitleEntryPtr entry = SubtitleEntryPtr(new SubtitleEntry());
				entry->SetStartTime(start);
				entry->SetEndTime(end);
				m_Entries.push_back(entry);
				String entryLine;
				int inside = 0;
				while (pos < (int) s.length()) {
					if (s[pos] == TEXT('{')) {
						inside++;
					} else if (s[pos] == TEXT('}')) {
						inside--;
					} else if (! inside) {
						if (s[pos] == TEXT('\\')) {
							pos++;
							if (pos < (int) s.length()) {
								if (s[pos] == TEXT('n') || s[pos] == TEXT('N')) {
									entry->AppendLine(entryLine);
									entryLine.clear();
								} else if (s[pos] == TEXT('b')) {
									entryLine += TEXT(' ');
								} else {
									entryLine += s[pos];
								}
							}
						} else {
							entryLine += s[pos];
						}
					}
					pos++;
				}
				if (entryLine.length()) {
					entry->AppendLine(entryLine);
				}
			}
		}
	}
}				

bool SubtitleReader::ReadSubtitles(const String & fileName, bool utf8, double fps, double offset, double acceleration)
{
	m_Utf8         = utf8;
	m_SourceFPS    = fps;
	m_Offset       = offset;
	m_Acceleration = acceleration;

	m_FoundSomething = false;
	m_State          = STATE_LOOKINGFORSOMETHING;
	m_LastEntry      = NULL;

	FILE * f = fopen(fileName.toAnsi().c_str(), "rt");
	if (! f) return false;
	char * line = NULL;
	while (line = safefgets(line, f)) {
		switch (m_State) 
		{
		case STATE_LOOKINGFORSOMETHING:
			LookForSomething(line);
			break;
		case STATE_SRT_INSIDEENTRY:
			FetchSrtLine(line);
			break;
		case STATE_SSA_FORMATEXPECTED:
			ProcessSsaFormat(line);
			break;
		case STATE_SSA_DIALOGUEEXPECTED:
			ProcessSsaDialogue(line);
			break;
		}
	}
	fclose(f);
	if (! m_FoundSomething) return false;
	
	if (m_Entries.size()) {
		std::sort(m_Entries.begin(), m_Entries.end());
		SubtitleEntryVector::iterator last = m_Entries.begin();
		SubtitleEntryVector::iterator next = last + 1;
		while (next != m_Entries.end()) {
			if ((*last)->GetEndTime() > (*next)->GetStartTime()) {
				SubtitleEntryVector::iterator inserted = m_Entries.insert(next, SubtitleEntryPtr(new SubtitleEntry()));
				last = inserted - 1;
				next = inserted + 1;
				(*inserted)->AppendLines(**last);
				(*inserted)->AppendLines(**next);
				(*inserted)->SetStartTime((*next)->GetStartTime());
				(*inserted)->SetEndTime((*last)->GetEndTime());
				(*last)->SetEndTime((*inserted)->GetStartTime());
				(*next)->SetStartTime((*inserted)->GetEndTime());
				if ((*last)->GetStartTime() >= (*last)->GetEndTime()) {
					inserted = m_Entries.erase(last);
					next = inserted + 1;
				}
				last = inserted;
				if ((*next)->GetStartTime() >= (*next)->GetEndTime()) {
					next = m_Entries.erase(next);
				} else {
					last++;
					next++;
				}
			} else {
				last++;
				next++;
			}
		}
	}

	return true;
}
