#include "video.h"
#include "group.h"
#include "project.h"
#include "tools.h"
#include "avswrapper.h"
#include "langcode.h"
#include "title.h"
#include "config.h"

#include "safefgets/safefgets.h"

#include <algorithm>

Video::Video(Group * group) : ManagablePtr()
{
	m_Group           = group;
	m_MergeWithNext   = false;
	m_MergeWithPrev   = false;
	m_AspectRatio     = -1;
	m_AspectNom       = 1;
	m_AspectDen       = 1;
	m_CompensateOverscan = g_Config->GetCompensateOverscan();
	m_OverscanPercent = g_Config->GetOverscanPercent();
	m_CreateChapters  = g_Config->GetCreateChapters();
	m_ChapterCount    = g_Config->GetChapterCount();
	m_UpsideDown      = false;
	m_Interlacing     = 0;
	m_DoIVTC          = false;
	m_Opener          = g_Config->GetVideoOpener();
	m_PreProcessor    = g_Config->GetVideoPreProcessor();
	m_Title           = NULL;
	m_TitleNo         = 0;
	m_PartNo          = 0;
	m_FileSelected    = false;
	m_TrackSelected   = false;
	m_TitleSelected   = false;
	m_Index           = -1;
	SetInvalid(true);
}

Video::~Video()
{
}

Group * Video::GetGroup() const
{
	return m_Group;
}

void Video::AddTracks(bool selectThem)
{
	m_Audios.clear();
	for (int i = 0; i < m_Group->GetAudioCount(); i++) {
		AudioPtr audio = new Audio(this);
		m_Audios.push_back(audio);
		if (i == 0 && HasAudio()) {
			audio->SetFromVideo(true, false);
		}
		if (selectThem) {
			audio->SetTrackSelected(true);
			audio->SetFileSelected(true);
		}
	}
	m_Subtitles.clear();
	for (int i = 0; i < m_Group->GetSubtitleCount(); i++) {
		SubtitlePtr subtitle = new Subtitle(this);
		m_Subtitles.push_back(subtitle);
		if (selectThem) {
			subtitle->SetTrackSelected(true);
			subtitle->SetFileSelected(true);
		}
	}
	Recalculate();
}

bool Video::IsFileSelected() const
{
	return m_FileSelected;
}


void Video::SetFileSelected(bool selected)
{
	m_FileSelected = selected;
}

bool Video::IsTrackSelected() const
{
	return m_TrackSelected;
}

void Video::SetTrackSelected(bool selected)
{
	m_TrackSelected = selected;
}

bool Video::IsTitleSelected() const
{
	return m_TitleSelected;
}

void Video::SetTitleSelected(bool selected)
{
	m_TitleSelected = selected;
}

bool Video::IsSelected() const
{
	return IsFileSelected() || 
		(GetTitle()->GetVideoCount() == 1 && GetTitle()->IsSelected());
}

const String & Video::GetFileName() const
{
	return m_FileName;
}

bool Video::SetFileName(const String & fileName, bool setChanged)
{
	String fullPath = fileName.toFullPath();
	if (m_FileName != fullPath) {
		m_FileName = fullPath;
		SetInvalid(true);
		if (setChanged) SetChanged();
	}
	return true;
}

bool Video::IsMergedWithNext() const
{
	return m_MergeWithNext;
}

bool Video::SetMergeWithNext(bool mergeWithNext, bool setChanged)
{
	if (m_MergeWithNext != mergeWithNext) {
		m_MergeWithNext = mergeWithNext;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Video::IsMergedWithPrev() const
{
	return m_MergeWithPrev;
}

void Video::SetMergeWithPrev(bool mergeWithPrev)
{
	m_MergeWithPrev = mergeWithPrev;
}

int  Video::GetAspectRatio() const
{
	return m_AspectRatio;
}

bool Video::SetAspectRatio(int aspectRatio, bool setChanged)
{
	if (0 <= aspectRatio && aspectRatio < ASPECTRATIO_MAX) {
		if (m_AspectRatio != aspectRatio) {
			m_AspectRatio = aspectRatio;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

int  Video::GetAspectNom() const
{
	return m_AspectNom;
}

bool Video::SetAspectNom(int aspectNom, bool setChanged)
{
	if (0 < aspectNom) {
		if (m_AspectNom != aspectNom) {
			m_AspectNom = aspectNom;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

int  Video::GetAspectDen() const
{
	return m_AspectDen;
}

bool Video::SetAspectDen(int aspectDen, bool setChanged)
{
	if (0 < aspectDen) {
		if (m_AspectDen != aspectDen) {
			m_AspectDen = aspectDen;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

double Video::GetOrigAspectRatio() const
{
	switch (m_AspectRatio) {
	case ASPECTRATIO_4_3:
		return 4.0 / 3.0;
	case ASPECTRATIO_16_9:
		return 16.0 / 9.0;
	case ASPECTRATIO_2_1:
		return 2.0;
	case ASPECTRATIO_239_1:
		return 2.39;
	case ASPECTRATIO_CUSTOM:
		return 1.0 * m_AspectNom / m_AspectDen;
	}
	return 1.0;
}

String Video::GetOrigAspectRatioStr() const
{
	switch (m_AspectRatio) {
	case ASPECTRATIO_4_3:
		return S("4:3");
	case ASPECTRATIO_16_9:
		return S("16:9");
	case ASPECTRATIO_2_1:
		return S("2:1");
	case ASPECTRATIO_239_1:
		return S("2.39:1");
	case ASPECTRATIO_CUSTOM:
		return String(m_AspectNom) + S(":") + String(m_AspectDen);
	}
	return 1.0;
}

double Video::GetFinalAspectRatio() const
{
	switch (m_AspectRatio) {
	case ASPECTRATIO_4_3:
		return 4.0 / 3.0;
	case ASPECTRATIO_16_9:
	case ASPECTRATIO_2_1:
	case ASPECTRATIO_239_1:
		return 16.0 / 9.0;
	case ASPECTRATIO_CUSTOM:
		if (1.0 * m_AspectNom / m_AspectDen < 16.0 / 9.0) {
			return 4.0 / 3.0;
		} else {
			return 16.0 / 9.0;
		}
	}
	return 1.0;
}

String Video::GetFinalAspectRatioStr() const
{
	switch (m_AspectRatio) {
	case ASPECTRATIO_4_3:
		return S("4:3");
	case ASPECTRATIO_16_9:
	case ASPECTRATIO_2_1:
	case ASPECTRATIO_239_1:
		return S("16:9");
	case ASPECTRATIO_CUSTOM:
		if (1.0 * m_AspectNom / m_AspectDen < 16.0 / 9.0) {
			return S("4:3");
		} else {
			return S("16:9");
		}
	}
	return 1.0;
}

bool Video::GetCompensateOverscan() const
{
	return m_CompensateOverscan;
}

bool Video::SetCompensateOverscan(bool compensateOverscan, bool setChanged)
{
	if (m_CompensateOverscan != compensateOverscan) {
		m_CompensateOverscan = compensateOverscan;
		if (setChanged) SetChanged();
	}
	return true;
}
		
int Video::GetOverscanPercent() const
{
	return m_OverscanPercent;
}

bool Video::SetOverscanPercent(int overscanPercent, bool setChanged)
{
	if (0 <= overscanPercent && overscanPercent < 100) {
		if (m_OverscanPercent != overscanPercent) {
			m_OverscanPercent = overscanPercent;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

bool Video::GetCreateChapters() const
{
	return m_CreateChapters;
}

bool Video::SetCreateChapters(bool createChapters, bool setChanged)
{
	if (m_CreateChapters != createChapters) {
		m_CreateChapters = createChapters;
		if (setChanged) SetChanged();
	}
	return true;
}

int  Video::GetChapterCount() const
{
	return m_ChapterCount;
}

bool Video::SetChapterCount(int chapterCount, bool setChanged)
{
	if (0 <= chapterCount) {
		if (m_ChapterCount != chapterCount) {
			m_ChapterCount = chapterCount;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

String Video::GetChaptersAt() const
{
	String result;
	for (std::vector<double>::const_iterator ci = m_ChaptersAt.begin(); ci != m_ChaptersAt.end(); ci++) {
		if (result.length()) result += S(", ");
		result += String().fromLength(*ci);
	}
	return result;
}

bool Video::SetChaptersAt(const String & chaptersAt, bool setChanged)
{
	if (chaptersAt != GetChaptersAt()) {
		m_ChaptersAt.clear();
		double chapterAt = 0.0;
		String::const_iterator si = chaptersAt.begin();
		while (ParseLength(si, chaptersAt.end(), chapterAt)) {
			m_ChaptersAt.push_back(chapterAt);
		}
		std::sort(m_ChaptersAt.begin(), m_ChaptersAt.end());
		bool first = true;
		double lastAt = 0.0;
		for (std::vector<double>::iterator ci = m_ChaptersAt.begin(); ci != m_ChaptersAt.end();) {
			if (! first && (*ci) - lastAt < 0.001) {
				ci = m_ChaptersAt.erase(ci);
			} else {
				lastAt = (*ci);
				ci++;
			}
			first = false;
		}
		if (setChanged) SetChanged();
	}
	return true;
}

const std::vector<double> & Video::GetChapterVector() const
{
	return m_ChaptersAt;
}

bool Video::GetUpsideDown() const
{
	return m_UpsideDown;
}

bool Video::SetUpsideDown(bool upsideDown, bool setChanged)
{
	if (m_UpsideDown != upsideDown) {
		m_UpsideDown = upsideDown;
		if (setChanged) SetChanged();
	}
	return true;
}

int Video::GetInterlacing() const
{
	return m_Interlacing;
}

bool Video::SetInterlacing(int interlacing, bool setChanged)
{
	if (0 <= interlacing && interlacing < INTERLACING_MAX) {
		if (m_Interlacing != interlacing) {
			m_Interlacing = interlacing;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

bool Video::GetDoIVTC() const
{
	if (GetInterlacing() <= INTERLACING_PROGRESSIVE) return false;
	return m_DoIVTC;
}

bool Video::SetDoIVTC(bool doIVTC, bool setChanged)
{
	if (m_DoIVTC != doIVTC) {
		m_DoIVTC = doIVTC;
		if (setChanged) SetChanged();
	}
	return true;
}

bool Video::GetFixedChapters() const
{
	return m_FixedChapters;
}

bool Video::SetFixedChapters(bool fixedChapters, bool setChanged)
{
	if (m_FixedChapters != fixedChapters) {
		m_FixedChapters = fixedChapters;
		if (setChanged) SetChanged();
	}
	return true;
}

String Video::GetOpener() const
{
	return (IsScript() ? S("") : m_Opener);
}

bool Video::SetOpener(const String & opener, bool setChanged)
{
	if (opener.length()) {
		if (m_Opener != opener) {
			m_Opener = opener;
			SetInvalid(true);
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

const String & Video::GetPreProcessor() const
{
	return m_PreProcessor;
}

bool Video::SetPreProcessor(const String & preProcessor, bool setChanged)
{
	if (m_PreProcessor != preProcessor) {
		m_PreProcessor = preProcessor;
		SetInvalid(true);
		if (setChanged) SetChanged();
	}
	return true;
}

void Video::SetChanged()
{
	m_Group->SetChanged();
}

int Video::GetAudioCount() const
{
	return (int) m_Audios.size();
}

AudioPtr Video::GetAudio(int index) const
{
	if (0 <= index && index < (int) m_Audios.size()) {
		return m_Audios[index];
	}
	return AudioPtr();
}

void Video::SetAudio(int index, AudioPtr audio)
{
	m_Audios[index] = audio;
}

void Video::AddAudio(bool selected)
{
	AudioPtr audio = new Audio(this);
	if (! m_Audios.size()) {
		if (HasAudio()) {
			audio->SetFromVideo(true, false);
			audio->Analyze();
		}
	}
	m_Audios.push_back(audio);
	audio->SetFileSelected(selected);
	audio->SetTrackSelected(selected);
	Recalculate();
}

void Video::MoveAudioUp(int index)
{
	AudioPtr audio = m_Audios[index];
	m_Audios[index] = m_Audios[index - 1];
	m_Audios[index - 1] = audio;
	Recalculate();
}

void Video::MoveAudioDown(int index)
{
	AudioPtr audio = m_Audios[index];
	m_Audios[index] = m_Audios[index + 1];
	m_Audios[index + 1] = audio;
	Recalculate();
}

void Video::DeleteAudio(int index)
{
	m_Audios.erase(m_Audios.begin() + index);
	Recalculate();
}

int Video::GetSubtitleCount() const
{
	return (int) m_Subtitles.size();
}

SubtitlePtr Video::GetSubtitle(int index) const
{
	if (0 <= index && index < (int) m_Subtitles.size()) {
		return m_Subtitles[index];
	}
	return SubtitlePtr();
}

void Video::SetSubtitle(int index, SubtitlePtr subtitle)
{
	m_Subtitles[index] = subtitle;
}

void Video::AddSubtitle(bool selected)
{
	SubtitlePtr subtitle = new Subtitle(this);
	m_Subtitles.push_back(subtitle);
	subtitle->SetFileSelected(selected);
	subtitle->SetTrackSelected(selected);
	subtitle->Recalculate();
	Recalculate();
}

void Video::MoveSubtitleUp(int index)
{
	SubtitlePtr subtitle = m_Subtitles[index];
	m_Subtitles[index] = m_Subtitles[index - 1];
	m_Subtitles[index - 1] = subtitle;
	Recalculate();
}

void Video::MoveSubtitleDown(int index)
{
	SubtitlePtr subtitle = m_Subtitles[index];
	m_Subtitles[index] = m_Subtitles[index + 1];
	m_Subtitles[index + 1] = subtitle;
	Recalculate();
}
void Video::DeleteSubtitle(int index)
{
	m_Subtitles.erase(m_Subtitles.begin() + index);
	Recalculate();
}

void Video::SetInvalid(bool clearChecked)
{
	if (clearChecked) m_IsChecked = false;
	m_Info         = AVSWrapper::Info();
	m_DoConvertFPS = false;
	m_ConvertToFPS = 0.0;
	m_AssumeFPS    = false;
	m_Length       = 0.0;
	for (int i = 0; i < (int) m_Audios.size(); i++) {
		if (m_Audios[i]->GetFromVideo()) {
			m_Audios[i]->SetInvalid(true);
		}
	}
}

bool Video::Analyze()
{
	m_IsChecked = true;
	SetInvalid(false);
	AVSWrapper avs;
	avs.Open(GetFileName(), GetOpener(), GetUpsideDown(), GetPreProcessor(), 
		(GetGroup()->GetProject()->GetDVDFormat() == DVDFORMAT_PAL));
	m_Info = avs.GetInfo();
	m_OpenError = avs.GetError();
	if (IsValid() && (m_AspectRatio < 0)) {
		double medianAspectRatio = ((4.0 / 3.0) + (16.0 / 9.0)) / 2;
		double calculatedAspectRatio = 1.0 * GetWidth() / GetHeight();
		if (calculatedAspectRatio > medianAspectRatio) {
			m_AspectRatio = ASPECTRATIO_16_9;
		} else {
			m_AspectRatio = ASPECTRATIO_4_3;
		}
	}
	Recalculate();
	return IsValid();
}

int Video::GetIndex() const
{
	return m_Index;
}

void Video::SetIndex(int index)
{
	m_Index = index;
}

String Video::GetIdent(bool forTrack)
{
	if (m_Index < 0) {
		return S("");
	} else if (forTrack) {
		return S("Video track ") + 
			S(" of title ") + String(GetTitle()->GetIndex() + 1) +
			S(" in group ") + String(GetTitle()->GetGroup()->GetIndex() + 1) +
			S(": ");
	} else {
		return S("Video file ") + String(GetIndex() + 1) +
			S(" of title ") + String(GetTitle()->GetIndex() + 1) +
			S(" in group ") + String(GetTitle()->GetGroup()->GetIndex() + 1) +
			S(": ");
	}
}

String Video::Recalculate()
{
	String error;
	if (! IsChecked()) Analyze();
	if (IsValid()) {
		double fpsRatio = 1.0;
		double frameRate = GetGroup()->GetProject()->GetFrameRate();
		double ivctAdjustment = (GetDoIVTC() ? 4.0 / 5.0 : 1.0);
		fpsRatio = GetOrigFPS() * ivctAdjustment / frameRate;
		if (fpsRatio >= (3.0 / 2.0)) {
			m_DoConvertFPS = true;
			bool go = true;
			int round = -16;
			while (go) {
				if (round <= 1) {
					m_ConvertToFPS = GetOrigFPS() * (3 - round) / (4 - round);
				} else {
					m_ConvertToFPS = GetOrigFPS() / round;
				}
				fpsRatio = m_ConvertToFPS * ivctAdjustment / frameRate;
				go = (fpsRatio >= (3.0 / 2.0)); 
				if (go) round++;
			}
		}
		m_AssumeFPS = (0.95 < fpsRatio && fpsRatio < 1.05);
		m_Length = GetOrigLength();
		if (m_AssumeFPS) {
			m_Length *= fpsRatio;
			m_Acceleration = fpsRatio;
		} else {
			m_Acceleration = 1.0;
		}
	} else {
		error += (GetIdent(false) + GetFileName().enclose() + S(" doesn't seem to be a video file."));
		if (m_OpenError.length()) error += S(" (") + m_OpenError + S(")");
		error += S("\n");
	}

	m_SubStreamRate = 0;
	if (GetIndex() >= 0) {
		for (int aindex = 0; aindex < GetAudioCount(); aindex++) {
			Audio * audio = GetAudio(aindex);
			audio->SetIndex(aindex);
			error += audio->Recalculate();
			m_SubStreamRate += audio->GetBitRateInKbps();
		}
		for (int sindex = 0; sindex < GetSubtitleCount(); sindex++) {
			Subtitle * subtitle = GetSubtitle(sindex);
			subtitle->SetIndex(sindex);
			error += subtitle->Recalculate();
			m_SubStreamRate += subtitle->GetBitRateInKbps();
		}
	}

	return error;
}

bool Video::IsChecked() const
{
	return m_IsChecked;
}

const AVSWrapper::Info & Video::GetInfo() const
{
	return m_Info;
}

bool Video::IsValid() const
{
	return GetInfo().hasVideo;
}

bool Video::IsScript() const
{
	return AVSWrapper::IsScript(m_FileName);
}

int Video::GetWidth() const
{
	return GetInfo().width;
}

int Video::GetHeight() const
{
	return GetInfo().height;
}

int Video::GetFrameCount() const
{
	return GetInfo().frameCount;
}

double Video::GetOrigFPS() const
{
	return GetInfo().frameRate;
}

bool Video::DoConvertFPS() const
{
	return m_DoConvertFPS;
}

double Video::ConvertToFPS() const
{
	return m_ConvertToFPS;
}

bool Video::AssumeFPS() const
{
	return m_AssumeFPS;
}

double Video::GetOrigLength() const
{
	if (IsValid()) {
		return 1.0 * GetInfo().frameCount / GetInfo().frameRate;
	} else {
		return 0.0;
	}
}

double Video::GetLength() const
{
	return m_Length;
}

double Video::GetAcceleration() const
{
	return m_Acceleration;
}

bool Video::IsInterlaced() const
{
	return (GetInterlacing() > INTERLACING_PROGRESSIVE && ! GetDoIVTC());
}

bool Video::IsBFF() const
{
	return (GetInterlacing() == INTERLACING_BFF);
}

bool Video::HasAudio() const
{
	return GetInfo().hasAudio;
}

int Video::GetNoOfChannels() const
{
	return GetInfo().noOfChannels;
}

int Video::GetSubStreamRate() const
{
	return m_SubStreamRate;
}

Title * Video::GetTitle() const
{
	return m_Title;
}

void Video::SetTitle(Title * title)
{
	m_Title = title;
}

void Video::SetTitleNo(int titleNo)
{
	m_TitleNo = titleNo;
}

int Video::GetTitleNo() const
{
	return m_TitleNo;
}

void Video::SetPartNo(int partNo)
{
	m_PartNo = partNo;
}

int Video::GetPartNo() const
{
	return m_PartNo;
}

void Video::GetDescription(StringVectorVector & strings, bool forTitle) const
{
	StringVector line;
	if (forTitle) {
		if (GetGroup()->GetView() == VIEW_TITLETRACKPART) {
			line.push_back(S("Video track: "));
		} else {
			line.push_back(S("Title ") + String(GetTitle()->GetIndex() + 1) + S(": "));
		}
		String s = String().fromLength(GetTitle()->GetLength());
		String origAspect = GetOrigAspectRatioStr();
		s += S(", ") + origAspect;
		String finalAspect = GetFinalAspectRatioStr();
		if (finalAspect != origAspect) {
			s += S("->") + finalAspect;
		}
		if (GetCompensateOverscan()) {
			s += (S(", ") + String(GetOverscanPercent()) + S("% overscan"));
		}
		if (GetCreateChapters()) {
			s += (S(", chapters every ") + String(GetChapterCount()) + S(" minutes"));
		}
		line.push_back(s);
		strings.push_back(line);
	} else {
		String s = S("Video file");
		if (GetTitle()->GetVideoCount() > 1) {
			s += S(" ") + String(GetIndex() + 1);
		}
		s += S(": ");
		line.push_back(s);
		s = GetFileName().toFileName();
		s += S(" (") + m_Opener + (m_PreProcessor.length() ? S("+") + m_PreProcessor : S("")) + S(")");
		line.push_back(s);
		strings.push_back(line);
		line.clear();
		if (GetGroup()->GetView() != VIEW_OUTLINE) {
			line.push_back(String());
			if (IsValid()) {
				s = String().fromLength(GetOrigLength()) 
					+ S(", ") + String(GetWidth()) + S("x") + String(GetHeight());
				if (GetUpsideDown()) {
					s += S(", upside down");
				}
				switch (GetInterlacing()) {
				case INTERLACING_AUTO:
					s += S(", auto-deinterlace");
					break;
				case INTERLACING_PROGRESSIVE:
					s += S(", progressive");
					break;
				case INTERLACING_TFF:
				case INTERLACING_BFF:
					s += S(", interlaced");
					s += (GetInterlacing() == INTERLACING_TFF ? S(" (tff)") : S(" (bff)"));
					if (GetDoIVTC()) {
						s += S(", inverse telecine");
					}
					break;
				}
				if (GetFixedChapters()) {
					s += S(", chapters at ") + GetChaptersAt();
				}
				line.push_back(s);
			} else {
				line.push_back(S("Error: ") + m_OpenError);
			}
			strings.push_back(line);
		}
	}
}
 
bool Video::Load(csv_file_t csv, int & lineno, int nest)
{
	bool ok = true;
	
	int field = 1;
	SetFileSelected(csv_get_int(csv, lineno, ++field, (m_FileSelected ? 1 : 0)) != 0);
	ok = ok && SetFileName(String().fromUtf8(csv_get_str(csv, lineno, ++field, m_FileName.toUtf8().c_str())), false);
	ok = ok && SetMergeWithNext(csv_get_int(csv, lineno, ++field, m_MergeWithNext ? 1 : 0) != 0, false);
	ok = ok && SetAspectRatio(csv_get_int(csv, lineno, ++field, m_AspectRatio), false);
	ok = ok && SetAspectNom(csv_get_int(csv, lineno, ++field, m_AspectNom), false);
	ok = ok && SetAspectDen(csv_get_int(csv, lineno, ++field, m_AspectDen), false);
	ok = ok && SetCreateChapters(csv_get_int(csv, lineno, ++field, m_CreateChapters ? 1 : 0) != 0, false);
	ok = ok && SetChapterCount(csv_get_int(csv, lineno, ++field, m_ChapterCount), false);
	ok = ok && SetChaptersAt(String().fromUtf8(csv_get_str(csv, lineno, ++field, GetChaptersAt().toUtf8().c_str())), false);
	ok = ok && SetUpsideDown(csv_get_int(csv, lineno, ++field, m_UpsideDown ? 1 : 0) != 0, false);
	++field;
	ok = ok && SetInterlacing(csv_get_int(csv, lineno, ++field, m_Interlacing), false);
	ok = ok && SetDoIVTC(csv_get_int(csv, lineno, ++field, m_DoIVTC ? 1 : 0) != 0, false);
	++field;
	++field;
	SetTitleSelected(csv_get_int(csv, lineno, ++field, (m_TitleSelected ? 1 : 0)) != 0);
	ok = ok && SetFixedChapters(csv_get_int(csv, lineno, ++field, m_FixedChapters ? 1 : 0) != 0, false);
	SetTrackSelected(csv_get_int(csv, lineno, ++field, (m_TrackSelected ? 1 : 0)) != 0);
	ok = ok && SetOpener(String().fromUtf8(csv_get_str(csv, lineno, ++field, m_Opener.toUtf8().c_str())), false);
	ok = ok && SetPreProcessor(String().fromUtf8(csv_get_str(csv, lineno, ++field, m_PreProcessor.toUtf8().c_str())), false);
	ok = ok && SetCompensateOverscan(csv_get_int(csv, lineno, ++field, m_CompensateOverscan ? 1 : 0) != 0, false);
	ok = ok && SetOverscanPercent(csv_get_int(csv, lineno, ++field, m_OverscanPercent), false);

	lineno++;
	while (ok && lineno < csv->count && csv_get_int(csv, lineno, 0, 0) == nest) {
		const char * type = csv_get_str(csv, lineno, 1, "");
		
		if (! strcmp(type, "audio")) {
			AudioPtr audio = new Audio(this);
			if (audio->Load(csv, lineno, nest + 1)) {
				m_Audios.push_back(audio);
			} else {
				ok = false;
			}
		
		} else if (! strcmp(type, "subtitle")) {
			SubtitlePtr subtitle = new Subtitle(this);
			if (subtitle->Load(csv, lineno, nest + 1)) {
				m_Subtitles.push_back(subtitle);
			} else {
				ok = false;
			}
		
		} else {
			lineno++;
		}
	}

	return ok;
}

void Video::Save(FILE * f, int nest) const
{
	String fileName  = String(GetAllocRelativePath(m_FileName.c_str(),  NULL), true);
	fprintf(f, "%d,video,%d,%s,%d,%d,%d,%d,%d,%d,%s,%d,%d,%d,%d,%d,%s,%d,%d,%d,%s,%s,%d,%d\n", nest,
		(m_FileSelected ? 1 : 0),
		fileName.enclose().toUtf8().c_str(),
		(m_MergeWithNext ? 1 : 0),
		m_AspectRatio,
		m_AspectNom,
		m_AspectDen,
		(m_CreateChapters ? 1 : 0),
		m_ChapterCount,
		GetChaptersAt().enclose().toUtf8().c_str(),
		(m_UpsideDown ? 1 : 0),
		0,
		m_Interlacing,
		(m_DoIVTC ? 1 : 0),
		0,
		"",
		(m_TitleSelected ? 1 : 0),
		(m_FixedChapters ? 1 : 0),
		(m_TrackSelected ? 1 : 0),
		m_Opener.enclose().toUtf8().c_str(),
		m_PreProcessor.enclose().toUtf8().c_str(),
		(m_CompensateOverscan ? 1 : 0),
		m_OverscanPercent);

	for (AudioVector::const_iterator audioIterator = m_Audios.begin();
		audioIterator != m_Audios.end(); audioIterator++) {
		(*audioIterator)->Save(f, nest + 1);
	}

	for (SubtitleVector::const_iterator subtitleIterator = m_Subtitles.begin();
		subtitleIterator != m_Subtitles.end(); subtitleIterator++) {
		(*subtitleIterator)->Save(f, nest + 1);
	}
}
